# Copyright 2021 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from configparser import ConfigParser

import nacl.encoding
import nacl.signing
import signedjson.key
import signedjson.types

from sydent.config._base import BaseConfig


class CryptoConfig(BaseConfig):
    def parse_config(self, cfg: "ConfigParser") -> bool:
        """
        Parse the crypto section of the config
        :param cfg: the configuration to be parsed
        """

        signing_key_str = cfg.get("crypto", "ed25519.signingkey")
        signing_key_parts = signing_key_str.split(" ")

        save_key = False

        # N.B. `signedjson` expects `nacl.signing.SigningKey` instances which
        # have been monkeypatched to include new `alg` and `version` attributes.
        # This is captured by the `signedjson.types.SigningKey` protocol.
        self.signing_key: signedjson.types.SigningKey

        if signing_key_str == "":
            print(
                "INFO: This server does not yet have an ed25519 signing key. "
                "Creating one and saving it in the config file."
            )

            self.signing_key = signedjson.key.generate_signing_key("0")

            save_key = True
        elif len(signing_key_parts) == 1:
            # old format key
            print("INFO: Updating signing key format: brace yourselves")

            self.signing_key = nacl.signing.SigningKey(
                signing_key_str.encode("ascii"), encoder=nacl.encoding.HexEncoder
            )
            self.signing_key.version = "0"
            self.signing_key.alg = signedjson.key.NACL_ED25519

            save_key = True
        else:
            self.signing_key = signedjson.key.decode_signing_key_base64(
                signing_key_parts[0], signing_key_parts[1], signing_key_parts[2]
            )

        if save_key:
            signing_key_str = "%s %s %s" % (
                self.signing_key.alg,
                self.signing_key.version,
                signedjson.key.encode_signing_key_base64(self.signing_key),
            )
            cfg.set("crypto", "ed25519.signingkey", signing_key_str)
            return True
        else:
            return False



# Copyright 2019 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import TYPE_CHECKING, Dict, List, Mapping, Optional, Set, Union

import yaml
from typing_extensions import TypedDict

if TYPE_CHECKING:
    from sydent.sydent import Sydent

logger = logging.getLogger(__name__)


class TermConfig(TypedDict):
    master_version: str
    docs: Mapping[str, "Policy"]


class Policy(TypedDict):
    version: str
    langs: Mapping[str, "LocalisedPolicy"]


class LocalisedPolicy(TypedDict):
    name: str
    url: str


VersionOrLang = Union[str, LocalisedPolicy]


class Terms:
    def __init__(self, yamlObj: Optional[TermConfig]) -> None:
        """
        :param yamlObj: The parsed YAML.
        """
        self._rawTerms = yamlObj

    def getMasterVersion(self) -> Optional[str]:
        """
        :return: The global (master) version of the terms, or None if there
            are no terms of service for this server.
        """
        if self._rawTerms is None:
            return None
        return self._rawTerms["master_version"]

    def getForClient(self) -> Dict[str, Dict[str, Dict[str, VersionOrLang]]]:
        # Examples:
        # "policy" -> "terms_of_service", "version" -> "1.2.3"
        # "policy" -> "terms_of_service", "en" -> LocalisedPolicy
        """
        :return: A dict which value for the "policies" key is a dict which contains the
            "docs" part of the terms' YAML. That nested dict is empty if no terms.
        """
        policies: Dict[str, Dict[str, VersionOrLang]] = {}
        if self._rawTerms is not None:
            for docName, doc in self._rawTerms["docs"].items():
                policies[docName] = {
                    "version": doc["version"],
                }
                policies[docName].update(doc["langs"])
        return {"policies": policies}

    def getUrlSet(self) -> Set[str]:
        """
        :return: All the URLs for the terms in a set. Empty set if no terms.
        """
        urls = set()
        if self._rawTerms is not None:
            for docName, doc in self._rawTerms["docs"].items():
                for langName, lang in doc["langs"].items():
                    url = lang["url"]
                    urls.add(url)
        return urls

    def urlListIsSufficient(self, urls: List[str]) -> bool:
        """
        Checks whether the provided list of URLs (which represents the list of terms
        accepted by the user) is enough to allow the creation of the user's account.

        :param urls: The list of URLs of terms the user has accepted.

        :return: Whether the list is sufficient to allow the creation of the user's
            account.
        """
        agreed = set()
        urlset = set(urls)

        if self._rawTerms is None:
            if urls:
                raise ValueError("No configured terms, but user accepted some terms")
            else:
                return True

        else:
            for docName, doc in self._rawTerms["docs"].items():
                for lang in doc["langs"].values():
                    if lang["url"] in urlset:
                        agreed.add(docName)
                        break

            required = set(self._rawTerms["docs"].keys())
            return agreed == required


def get_terms(sydent: "Sydent") -> Terms:
    """Read and parse terms as specified in the config.

    Errors in reading, parsing and validating the config
    are raised as exceptions."""
    # TODO - move some of this to parse_config

    termsPath = sydent.config.general.terms_path

    if termsPath == "":
        return Terms(None)

    with open(termsPath) as fp:
        termsYaml = yaml.safe_load(fp)

    # TODO use something like jsonschema instead of this handwritten code.
    if "master_version" not in termsYaml:
        raise Exception("No master version")
    elif not isinstance(termsYaml["master_version"], str):
        raise TypeError(
            f"master_version should be a string, not {termsYaml['master_version']!r}"
        )
    if "docs" not in termsYaml:
        raise Exception("No 'docs' key in terms")
    for docName, doc in termsYaml["docs"].items():
        if "version" not in doc:
            raise Exception("'%s' has no version" % (docName,))
        if "langs" not in doc:
            raise Exception("'%s' has no langs" % (docName,))
        for langKey, lang in doc["langs"].items():
            if "name" not in lang:
                raise Exception("lang '%s' of doc %s has no name" % (langKey, docName))
            if "url" not in lang:
                raise Exception("lang '%s' of doc %s has no url" % (langKey, docName))

    return Terms(termsYaml)



# Copyright 2014 OpenMarket Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Any, Dict

import signedjson.sign

if TYPE_CHECKING:
    from sydent.sydent import Sydent
    from sydent.threepid import ThreepidAssociation


class Signer:
    def __init__(self, sydent: "Sydent") -> None:
        self.sydent = sydent

    def signedThreePidAssociation(self, assoc: "ThreepidAssociation") -> Dict[str, Any]:
        """
        Signs a 3PID association.

        :param assoc: The association to sign.

        :return: A signed representation of the association.
        """
        sgassoc = {
            "medium": assoc.medium,
            "address": assoc.address,
            "mxid": assoc.mxid,
            "ts": assoc.ts,
            "not_before": assoc.not_before,
            "not_after": assoc.not_after,
        }
        sgassoc.update(assoc.extra_fields)
        sgassoc = signedjson.sign.sign_json(
            sgassoc, self.sydent.config.general.server_name, self.sydent.keyring.ed25519
        )
        return sgassoc



# Copyright 2014 OpenMarket Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

from twisted.web.server import Request

from sydent.db.valsession import ThreePidValSessionStore
from sydent.http.auth import authV2
from sydent.http.servlets import SydentResource, get_args, jsonwrap, send_cors
from sydent.types import JsonDict
from sydent.util.stringutils import is_valid_client_secret
from sydent.validators import (
    IncorrectClientSecretException,
    InvalidSessionIdException,
    SessionExpiredException,
    SessionNotValidatedException,
)

if TYPE_CHECKING:
    from sydent.sydent import Sydent


class GetValidated3pidServlet(SydentResource):
    isLeaf = True

    def __init__(self, syd: "Sydent", require_auth: bool = False) -> None:
        super().__init__()
        self.sydent = syd
        self.require_auth = require_auth

    @jsonwrap
    def render_GET(self, request: Request) -> JsonDict:
        send_cors(request)
        if self.require_auth:
            authV2(self.sydent, request)

        args = get_args(request, ("sid", "client_secret"))

        sid = args["sid"]
        clientSecret = args["client_secret"]

        if not is_valid_client_secret(clientSecret):
            request.setResponseCode(400)
            return {
                "errcode": "M_INVALID_PARAM",
                "error": "Invalid client_secret provided",
            }

        valSessionStore = ThreePidValSessionStore(self.sydent)

        noMatchError = {
            "errcode": "M_NO_VALID_SESSION",
            "error": "No valid session was found matching that sid and client secret",
        }

        try:
            s = valSessionStore.getValidatedSession(sid, clientSecret)
        except (IncorrectClientSecretException, InvalidSessionIdException):
            request.setResponseCode(404)
            return noMatchError
        except SessionExpiredException:
            request.setResponseCode(400)
            return {
                "errcode": "M_SESSION_EXPIRED",
                "error": "This validation session has expired: call requestToken again",
            }
        except SessionNotValidatedException:
            request.setResponseCode(400)
            return {
                "errcode": "M_SESSION_NOT_VALIDATED",
                "error": "This validation session has not yet been completed",
            }

        return {"medium": s.medium, "address": s.address, "validated_at": s.mtime}

    def render_OPTIONS(self, request: Request) -> bytes:
        send_cors(request)
        return b""



# Copyright 2021 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from configparser import ConfigParser

from sydent.config._base import BaseConfig


class DatabaseConfig(BaseConfig):
    def parse_config(self, cfg: "ConfigParser") -> bool:
        """
        Parse the database section of the config

        :param cfg: the configuration to be parsed
        """
        self.database_path = cfg.get("db", "db.file")

        return False



# Copyright 2021 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from configparser import ConfigParser


class BaseConfig(ABC):
    @abstractmethod
    def parse_config(self, cfg: ConfigParser) -> bool:
        """
        Parse the a section of the config

        :param cfg: the configuration to be parsed

        :return: whether or not cfg has been altered. This method CAN
            return True, but it *shouldn't* as this leads to altering the
            config file.
        """
        pass



import json
import logging
import os
from io import BytesIO
from typing import Dict, Optional
from unittest.mock import MagicMock

import attr
import twisted.logger
from OpenSSL import crypto
from twisted.internet import address
from twisted.internet._resolver import SimpleResolverComplexifier
from twisted.internet.defer import fail, succeed
from twisted.internet.error import DNSLookupError
from twisted.internet.interfaces import (
    IHostnameResolver,
    IReactorPluggableNameResolver,
    IResolverSimple,
)
from twisted.test.proto_helpers import MemoryReactorClock
from twisted.web.http import unquote
from twisted.web.http_headers import Headers
from twisted.web.server import Request, Site
from zope.interface import implementer

from sydent.config import SydentConfig
from sydent.sydent import Sydent

# Expires on Jan 11 2030 at 17:53:40 GMT
FAKE_SERVER_CERT_PEM = """
-----BEGIN CERTIFICATE-----
MIIDlzCCAn+gAwIBAgIUC8tnJVZ8Cawh5tqr7PCAOfvyGTYwDQYJKoZIhvcNAQEL
BQAwWzELMAkGA1UEBhMCQVUxEzARBgNVBAgMClNvbWUtU3RhdGUxITAfBgNVBAoM
GEludGVybmV0IFdpZGdpdHMgUHR5IEx0ZDEUMBIGA1UEAwwLZmFrZS5zZXJ2ZXIw
HhcNMjAwMTE0MTc1MzQwWhcNMzAwMTExMTc1MzQwWjBbMQswCQYDVQQGEwJBVTET
MBEGA1UECAwKU29tZS1TdGF0ZTEhMB8GA1UECgwYSW50ZXJuZXQgV2lkZ2l0cyBQ
dHkgTHRkMRQwEgYDVQQDDAtmYWtlLnNlcnZlcjCCASIwDQYJKoZIhvcNAQEBBQAD
ggEPADCCAQoCggEBANNzY7YHBLm4uj52ojQc/dfQCoR+63IgjxZ6QdnThhIlOYgE
3y0Ks49bt3GKmAweOFRRKfDhJRKCYfqZTYudMcdsQg696s2HhiTY0SpqO0soXwW4
6kEIxnTy2TqkPjWlsWgGTtbVnKc5pnLs7MaQwLIQfxirqD2znn+9r68WMOJRlzkv
VmrXDXjxKPANJJ9b0PiGrL2SF4QcF3zHk8Tjf24OGRX4JTNwiGraU/VN9rrqSHug
CLWcfZ1mvcav3scvtGfgm4kxcw8K6heiQAc3QAMWIrdWhiunaWpQYgw7euS8lZ/O
C7HZ7YbdoldknWdK8o7HJZmxUP9yW9Pqa3n8p9UCAwEAAaNTMFEwHQYDVR0OBBYE
FHwfTq0Mdk9YKqjyfdYm4v9zRP8nMB8GA1UdIwQYMBaAFHwfTq0Mdk9YKqjyfdYm
4v9zRP8nMA8GA1UdEwEB/wQFMAMBAf8wDQYJKoZIhvcNAQELBQADggEBAEPVM5/+
Sj9P/CvNG7F2PxlDQC1/+aVl6ARAz/bZmm7yJnWEleBSwwFLerEQU6KFrgjA243L
qgY6Qf2EYUn1O9jroDg/IumlcQU1H4DXZ03YLKS2bXFGj630Piao547/l4/PaKOP
wSvwDcJlBatKfwjMVl3Al/EcAgUJL8eVosnqHDSINdBuFEc8Kw4LnDSFoTEIx19i
c+DKmtnJNI68wNydLJ3lhSaj4pmsX4PsRqsRzw+jgkPXIG1oGlUDMO3k7UwxfYKR
XkU5mFYkohPTgxv5oYGq2FCOPixkbov7geCEvEUs8m8c8MAm4ErBUzemOAj8KVhE
tWVEpHfT+G7AjA8=
-----END CERTIFICATE-----
"""


def make_sydent(test_config: Optional[dict] = None) -> Sydent:
    """Create a new sydent

    Args:
        test_config: Configuration variables for overriding the default sydent
            config
    """
    if test_config is None:
        test_config = {}

    # Use an in-memory SQLite database. Note that the database isn't cleaned up between
    # tests, so by default the same database will be used for each test if changed to be
    # a file on disk.
    test_config.setdefault("db", {}).setdefault("db.file", ":memory:")

    # Specify a server name to avoid warnings.
    general_config = test_config.setdefault("general", {})
    general_config.setdefault("server.name", ":test:")
    # Specify the default templates.
    general_config.setdefault(
        "templates.path",
        os.path.join(os.path.dirname(os.path.dirname(__file__)), "res"),
    )

    # Specify a signing key.
    test_config.setdefault("crypto", {}).setdefault(
        "ed25519.signingkey", "ed25519 0 FJi1Rnpj3/otydngacrwddFvwz/dTDsBv62uZDN2fZM"
    )

    reactor = ResolvingMemoryReactorClock()

    sydent_config = SydentConfig()
    sydent_config.parse_config_dict(test_config)

    return Sydent(
        reactor=reactor,
        sydent_config=sydent_config,
        use_tls_for_federation=False,
    )


@attr.s
class FakeChannel:
    """
    A fake Twisted Web Channel (the part that interfaces with the
    wire). Mostly copied from Synapse's tests framework.
    """

    site = attr.ib(type=Site)
    _reactor = attr.ib()
    result = attr.ib(default=attr.Factory(dict))
    _producer = None

    @property
    def json_body(self):
        if not self.result:
            raise Exception("No result yet.")
        return json.loads(self.result["body"].decode("utf8"))

    @property
    def code(self):
        if not self.result:
            raise Exception("No result yet.")
        return int(self.result["code"])

    @property
    def headers(self):
        if not self.result:
            raise Exception("No result yet.")
        h = Headers()
        for i in self.result["headers"]:
            h.addRawHeader(*i)
        return h

    def writeHeaders(self, version, code, reason, headers):
        self.result["version"] = version
        self.result["code"] = code
        self.result["reason"] = reason
        self.result["headers"] = headers

    def write(self, content):
        assert isinstance(content, bytes), "Should be bytes! " + repr(content)

        if "body" not in self.result:
            self.result["body"] = b""

        self.result["body"] += content

    def registerProducer(self, producer, streaming):
        self._producer = producer
        self.producerStreaming = streaming

        def _produce():
            if self._producer:
                self._producer.resumeProducing()
                self._reactor.callLater(0.1, _produce)

        if not streaming:
            self._reactor.callLater(0.0, _produce)

    def unregisterProducer(self):
        if self._producer is None:
            return

        self._producer = None

    def requestDone(self, _self):
        self.result["done"] = True

    def getPeer(self):
        # We give an address so that getClientAddress().host returns a non null entry,
        # causing us to record the MAU
        return address.IPv4Address("TCP", "127.0.0.1", 3423)

    def getHost(self):
        return None

    @property
    def transport(self):
        return self

    def getPeerCertificate(self):
        """Returns the hardcoded TLS certificate for fake.server."""
        return crypto.load_certificate(crypto.FILETYPE_PEM, FAKE_SERVER_CERT_PEM)


class FakeSite:
    """A fake Twisted Web Site."""

    pass


def make_request(
    reactor,
    site,
    method,
    path,
    content=b"",
    access_token=None,
    request=Request,
    shorthand=True,
    federation_auth_origin=None,
):
    """
    Make a web request using the given method and path, feed it the
    content, and return the Request and the Channel underneath. Mostly

    Args:
        reactor (IReactor): The Twisted reactor to use when performing the request.
        site (
        method (bytes or unicode): The HTTP request method ("verb").
        path (bytes or unicode): The HTTP path, suitably URL encoded (e.g.
        escaped UTF-8 & spaces and such).
        content (bytes or dict): The body of the request. JSON-encoded, if
        a dict.
        access_token (unicode): An access token to use to authenticate the request,
            None if no access token needs to be included.
        request (IRequest): The class to use when instantiating the request object.
        shorthand: Whether to try and be helpful and prefix the given URL
        with the usual REST API path, if it doesn't contain it.
        federation_auth_origin (bytes|None): if set to not-None, we will add a fake
            Authorization header pretenting to be the given server name.

    Returns:
        Tuple[synapse.http.site.SynapseRequest, channel]
    """
    if not isinstance(method, bytes):
        method = method.encode("ascii")

    if not isinstance(path, bytes):
        path = path.encode("ascii")

    # Decorate it to be the full path, if we're using shorthand
    if shorthand and not path.startswith(b"/_matrix"):
        path = b"/_matrix/identity/v2/" + path
        path = path.replace(b"//", b"/")

    if not path.startswith(b"/"):
        path = b"/" + path

    if isinstance(content, dict):
        content = json.dumps(content)
    if isinstance(content, str):
        content = content.encode("utf8")

    channel = FakeChannel(site, reactor)

    req = request(channel)
    req.content = BytesIO(content)
    req.postpath = list(map(unquote, path[1:].split(b"/")))

    if access_token:
        req.requestHeaders.addRawHeader(
            b"Authorization", b"Bearer " + access_token.encode("ascii")
        )

    if federation_auth_origin is not None:
        req.requestHeaders.addRawHeader(
            b"Authorization",
            b"X-Matrix origin=%s,key=,sig=" % (federation_auth_origin,),
        )

    if content:
        req.requestHeaders.addRawHeader(b"Content-Type", b"application/json")

    req.requestReceived(method, path, b"1.1")

    return req, channel


class ToTwistedHandler(logging.Handler):
    """logging handler which sends the logs to the twisted log"""

    tx_log = twisted.logger.Logger()

    def emit(self, record):
        log_entry = self.format(record)
        log_level = record.levelname.lower().replace("warning", "warn")
        self.tx_log.emit(
            twisted.logger.LogLevel.levelWithName(log_level), "{entry}", entry=log_entry
        )


def setup_logging():
    """Configure the python logging appropriately for the tests.

    (Logs will end up in _trial_temp.)
    """
    root_logger = logging.getLogger()

    log_format = "%(asctime)s - %(name)s - %(lineno)d - %(levelname)s" " - %(message)s"

    handler = ToTwistedHandler()
    formatter = logging.Formatter(log_format)
    handler.setFormatter(formatter)
    root_logger.addHandler(handler)

    log_level = os.environ.get("SYDENT_TEST_LOG_LEVEL", "ERROR")
    root_logger.setLevel(log_level)


setup_logging()


@implementer(IReactorPluggableNameResolver)
class ResolvingMemoryReactorClock(MemoryReactorClock):
    """
    A MemoryReactorClock that supports name resolution.
    """

    def __init__(self):
        lookups = self.lookups = {}  # type: Dict[str, str]

        @implementer(IResolverSimple)
        class FakeResolver:
            def getHostByName(self, name, timeout=None):
                if name not in lookups:
                    return fail(DNSLookupError("OH NO: unknown %s" % (name,)))
                return succeed(lookups[name])

        self.nameResolver = SimpleResolverComplexifier(FakeResolver())
        super().__init__()

    def installNameResolver(self, resolver: IHostnameResolver) -> IHostnameResolver:
        raise NotImplementedError()


class AsyncMock(MagicMock):
    async def __call__(self, *args, **kwargs):
        return super().__call__(*args, **kwargs)



from unittest.mock import Mock, patch

from twisted.trial import unittest
from twisted.web.client import Response

from sydent.db.invite_tokens import JoinTokenStore
from sydent.http.httpclient import FederationHttpClient
from sydent.http.servlets.store_invite_servlet import StoreInviteServlet
from tests.utils import make_request, make_sydent


class ThreepidInvitesTestCase(unittest.TestCase):
    """Tests features related to storing and delivering 3PID invites."""

    def setUp(self):
        # Create a new sydent
        config = {
            "email": {
                # Used by test_invited_email_address_obfuscation
                "email.third_party_invite_username_obfuscate_characters": "6",
                "email.third_party_invite_domain_obfuscate_characters": "8",
                "email.third_party_invite_keyword_blocklist": "evil\nbad\nhttps://",
            },
        }
        self.sydent = make_sydent(test_config=config)

    def test_delete_on_bind(self):
        """Tests that 3PID invite tokens are deleted upon delivery after a successful
        bind.
        """
        self.sydent.run()

        # The 3PID we're working with.
        medium = "email"
        address = "john@example.com"

        # Mock post_json_get_nothing so the /onBind call doesn't fail.
        async def post_json_get_nothing(uri, post_json, opts):
            return Response((b"HTTP", 1, 1), 200, b"OK", None, None)

        FederationHttpClient.post_json_get_nothing = Mock(
            side_effect=post_json_get_nothing,
        )

        # Manually insert an invite token, we'll check later that it's been deleted.
        join_token_store = JoinTokenStore(self.sydent)
        join_token_store.storeToken(
            medium,
            address,
            "!someroom:example.com",
            "@jane:example.com",
            "sometoken",
        )

        # Make sure the token still exists and can be retrieved.
        tokens = join_token_store.getTokens(medium, address)
        self.assertEqual(len(tokens), 1, tokens)

        # Bind the 3PID
        self.sydent.threepidBinder.addBinding(
            medium,
            address,
            "@john:example.com",
        )

        # Give Sydent some time to call /onBind and delete the token.
        self.sydent.reactor.advance(1000)

        cur = self.sydent.db.cursor()

        # Manually retrieve the tokens for this 3PID. We don't use getTokens because it
        # filters out sent tokens, so would return nothing even if the token hasn't been
        # deleted.
        res = cur.execute(
            "SELECT medium, address, room_id, sender, token FROM invite_tokens"
            " WHERE medium = ? AND address = ?",
            (
                medium,
                address,
            ),
        )
        rows = res.fetchall()

        # Check that we didn't get any result.
        self.assertEqual(len(rows), 0, rows)

    def test_invited_email_address_obfuscation(self):
        """Test that email addresses included in third-party invites are properly
        obfuscated according to the relevant config options
        """
        store_invite_servlet = StoreInviteServlet(self.sydent)

        email_address = "1234567890@1234567890.com"
        redacted_address = store_invite_servlet.redact_email_address(email_address)

        self.assertEqual(redacted_address, "123456...@12345678...")

        # Even short addresses are redacted
        short_email_address = "1@1.com"
        redacted_address = store_invite_servlet.redact_email_address(
            short_email_address
        )

        self.assertEqual(redacted_address, "...@1...")

    def test_third_party_invite_keyword_block_works(self):
        invite_config = {
            "medium": "email",
            "address": "foo@example.com",
            "room_id": "!bar",
            "sender": "@foo:example.com",
            "room_name": "This is an EVIL room name.",
        }
        request, channel = make_request(
            self.sydent.reactor,
            self.sydent.clientApiHttpServer.factory,
            "POST",
            "/_matrix/identity/api/v1/store-invite",
            invite_config,
        )
        self.assertEqual(channel.code, 403)

    def test_third_party_invite_keyword_blocklist_exempts_web_client_location_url(
        self,
    ):
        invite_config = {
            "medium": "email",
            "address": "foo@example.com",
            "room_id": "!bar",
            "sender": "@foo:example.com",
            "room_name": "This is a fine room name.",
            "org.matrix.web_client_location": "https://example.com",
        }

        # don't actually send the email
        with patch("sydent.util.emailutils.smtplib") as smtplib:
            request, channel = make_request(
                self.sydent.reactor,
                self.sydent.clientApiHttpServer.factory,
                "POST",
                "/_matrix/identity/api/v1/store-invite",
                invite_config,
            )
        self.assertEqual(channel.code, 200)
        smtp = smtplib.SMTP.return_value
        # but make sure we did try to send it
        smtp.sendmail.assert_called_once()


class ThreepidInvitesNoDeleteTestCase(unittest.TestCase):
    """Test that invite tokens are not deleted when that is disabled."""

    def setUp(self):
        # Create a new sydent
        config = {"general": {"delete_tokens_on_bind": "false"}}
        self.sydent = make_sydent(test_config=config)

    def test_no_delete_on_bind(self):
        self.sydent.run()

        # The 3PID we're working with.
        medium = "email"
        address = "john@example.com"

        # Mock post_json_get_nothing so the /onBind call doesn't fail.
        async def post_json_get_nothing(uri, post_json, opts):
            return Response((b"HTTP", 1, 1), 200, b"OK", None, None)

        FederationHttpClient.post_json_get_nothing = Mock(
            side_effect=post_json_get_nothing,
        )

        # Manually insert an invite token, we'll check later that it's been deleted.
        join_token_store = JoinTokenStore(self.sydent)
        join_token_store.storeToken(
            medium,
            address,
            "!someroom:example.com",
            "@jane:example.com",
            "sometoken",
        )

        # Make sure the token still exists and can be retrieved.
        tokens = join_token_store.getTokens(medium, address)
        self.assertEqual(len(tokens), 1, tokens)

        # Bind the 3PID
        self.sydent.threepidBinder.addBinding(
            medium,
            address,
            "@john:example.com",
        )

        # Give Sydent some time to call /onBind and delete the token.
        self.sydent.reactor.advance(1000)

        cur = self.sydent.db.cursor()

        # Manually retrieve the tokens for this 3PID. We don't use getTokens because it
        # filters out sent tokens, so would return nothing even if the token hasn't been
        # deleted.
        res = cur.execute(
            "SELECT medium, address, room_id, sender, token FROM invite_tokens"
            " WHERE medium = ? AND address = ?",
            (
                medium,
                address,
            ),
        )
        rows = res.fetchall()

        # Check that we didn't get any result.
        self.assertEqual(len(rows), 1, rows)



# Copyright 2021 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import socket
from configparser import ConfigParser
from typing import Optional

from sydent.config._base import BaseConfig
from sydent.config.exceptions import ConfigError
from sydent.util.emailutils import EmailAddressException, check_valid_email_address


class EmailConfig(BaseConfig):
    def parse_config(self, cfg: ConfigParser) -> bool:
        """
        Parse the email section of the config

        :param cfg: the configuration to be parsed
        """

        # These two options are deprecated
        self.template: Optional[str] = cfg.get("email", "email.template", fallback=None)

        self.invite_template = cfg.get("email", "email.invite_template", fallback=None)

        # This isn't used anywhere...
        self.validation_subject = cfg.get("email", "email.subject")

        self.invite_subject = cfg.get("email", "email.invite.subject", raw=True)
        self.invite_subject_space = cfg.get(
            "email", "email.invite.subject_space", raw=True
        )

        self.smtp_server = cfg.get("email", "email.smtphost")
        self.smtp_port = cfg.get("email", "email.smtpport")
        self.smtp_username = cfg.get("email", "email.smtpusername")
        self.smtp_password = cfg.get("email", "email.smtppassword")
        self.tls_mode = cfg.get("email", "email.tlsmode")

        # This is the fully qualified domain name for SMTP HELO/EHLO
        self.host_name = cfg.get("email", "email.hostname")
        if self.host_name == "":
            self.host_name = socket.getfqdn()

        self.sender = cfg.get("email", "email.from")
        try:
            check_valid_email_address(self.sender, allow_description=True)
        except EmailAddressException as e:
            raise ConfigError(f"Invalid email address '{self.sender}'") from e

        self.default_web_client_location = cfg.get(
            "email", "email.default_web_client_location"
        )

        self.username_obfuscate_characters = cfg.getint(
            "email", "email.third_party_invite_username_obfuscate_characters"
        )

        self.domain_obfuscate_characters = cfg.getint(
            "email", "email.third_party_invite_domain_obfuscate_characters"
        )

        third_party_invite_homeserver_blocklist = cfg.get(
            "email", "email.third_party_invite_homeserver_blocklist", fallback=""
        )
        third_party_invite_room_blocklist = cfg.get(
            "email", "email.third_party_invite_room_blocklist", fallback=""
        )
        third_party_invite_keyword_blocklist = cfg.get(
            "email", "email.third_party_invite_keyword_blocklist", fallback=""
        )
        self.third_party_invite_homeserver_blocklist = {
            server
            for server in third_party_invite_homeserver_blocklist.split("\n")
            if server  # filter out empty lines
        }
        self.third_party_invite_room_blocklist = {
            room_id
            for room_id in third_party_invite_room_blocklist.split("\n")
            if room_id  # filter out empty lines
        }
        self.third_party_invite_keyword_blocklist = {
            keyword.casefold()
            for keyword in third_party_invite_keyword_blocklist.split("\n")
            if keyword
        }

        self.email_sender_ratelimit_burst = cfg.getint(
            "email", "email.ratelimit_sender.burst", fallback=5
        )
        self.email_sender_ratelimit_rate_hz = cfg.getfloat(
            "email", "email.ratelimit_sender.rate_hz", fallback=1.0 / (5 * 60.0)
        )

        return False



#  Copyright 2021 The Matrix.org Foundation C.I.C.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import asyncio
import os.path
from typing import Optional
from unittest.mock import Mock, patch

import attr
from twisted.trial import unittest

from sydent.types import JsonDict
from tests.utils import make_request, make_sydent


@attr.s(auto_attribs=True)
class FakeHeader:
    """
    A fake header object
    """

    headers: dict

    def getAllRawHeaders(self):
        return self.headers


@attr.s(auto_attribs=True)
class FakeResponse:
    """A fake twisted.web.IResponse object"""

    # HTTP response code
    code: int

    # Fake Header
    headers: FakeHeader


class TestRequestCode(unittest.TestCase):
    def setUp(self) -> None:
        # Create a new sydent
        config = {
            "general": {
                "templates.path": os.path.join(
                    os.path.dirname(os.path.dirname(__file__)), "res"
                ),
            },
        }
        self.sydent = make_sydent(test_config=config)

    def _make_request(self, url: str, body: Optional[JsonDict] = None) -> Mock:
        # Patch out the email sending so we can investigate the resulting email.
        with patch("sydent.sms.openmarket.OpenMarketSMS.sendTextSMS") as sendTextSMS:
            # We can't use AsyncMock until Python 3.8. Instead, mock the
            # function as returning a future.
            f = asyncio.Future()
            f.set_result(Mock())
            sendTextSMS.return_value = f

            request, channel = make_request(
                self.sydent.reactor,
                self.sydent.clientApiHttpServer.factory,
                "POST",
                url,
                body,
            )
            self.assertEqual(channel.code, 200)

        return sendTextSMS

    def test_request_code(self) -> None:
        self.sydent.run()

        sendSMS_mock = self._make_request(
            "/_matrix/identity/api/v1/validate/msisdn/requestToken",
            {
                "phone_number": "447700900750",
                "country": "GB",
                "client_secret": "oursecret",
                "send_attempt": 0,
            },
        )
        sendSMS_mock.assert_called_once()

    def test_request_code_via_url_query_params(self) -> None:
        self.sydent.run()
        url = (
            "/_matrix/identity/api/v1/validate/msisdn/requestToken?"
            "phone_number=447700900750"
            "&country=GB"
            "&client_secret=oursecret"
            "&send_attempt=0"
        )
        sendSMS_mock = self._make_request(url)
        sendSMS_mock.assert_called_once()

    @patch("sydent.http.httpclient.HTTPClient.post_json_maybe_get_json")
    def test_bad_api_response_raises_exception(self, post_json: Mock) -> None:
        """Test that an error response from OpenMarket raises an exception
        and that the requester receives an error code."""

        header = FakeHeader({})
        resp = FakeResponse(code=400, headers=header), {}
        post_json.return_value = resp
        self.sydent.run()
        request, channel = make_request(
            self.sydent.reactor,
            self.sydent.clientApiHttpServer.factory,
            "POST",
            "/_matrix/identity/api/v1/validate/msisdn/requestToken",
            {
                "phone_number": "447700900750",
                "country": "GB",
                "client_secret": "oursecret",
                "send_attempt": 0,
            },
        )
        self.assertEqual(channel.code, 500)



# Copyright 2020 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from typing import Optional, Tuple

from twisted.internet.abstract import isIPAddress, isIPv6Address

# https://matrix.org/docs/spec/client_server/r0.6.0#post-matrix-client-r0-register-email-requesttoken
CLIENT_SECRET_REGEX = re.compile(r"^[0-9a-zA-Z\.=_\-]+$")

# hostname/domain name
# https://regex101.com/r/OyN1lg/2
hostname_regex = re.compile(
    r"^(?:[a-z0-9](?:[a-z0-9-]{0,61}[a-z0-9])?)(?:\.[a-z0-9](?:[a-z0-9-]{0,61}[a-z0-9])?)*$",
    flags=re.IGNORECASE,
)

# it's unclear what the maximum length of an email address is. RFC3696 (as corrected
# by errata) says:
#    the upper limit on address lengths should normally be considered to be 254.
#
# In practice, mail servers appear to be more tolerant and allow 400 characters
# or so. Let's allow 500, which should be plenty for everyone.
#
MAX_EMAIL_ADDRESS_LENGTH = 500


def is_valid_client_secret(client_secret: str) -> bool:
    """Validate that a given string matches the client_secret regex defined by the spec

    :param client_secret: The client_secret to validate

    :return: Whether the client_secret is valid
    """
    return (
        0 < len(client_secret) <= 255
        and CLIENT_SECRET_REGEX.match(client_secret) is not None
    )


def is_valid_hostname(string: str) -> bool:
    """Validate that a given string is a valid hostname or domain name.

    For domain names, this only validates that the form is right (for
    instance, it doesn't check that the TLD is valid).

    :param string: The string to validate

    :return: Whether the input is a valid hostname
    """

    return hostname_regex.match(string) is not None


def parse_server_name(server_name: str) -> Tuple[str, Optional[str]]:
    """Split a server name into host/port parts.

    No validation is done on the host part. The port part is validated to be
    a valid port number.

    Args:
        server_name: server name to parse

    Returns:
        host/port parts.

    Raises:
        ValueError if the server name could not be parsed.
    """
    try:
        if server_name[-1] == "]":
            # ipv6 literal, hopefully
            return server_name, None

        host_port = server_name.rsplit(":", 1)
        host = host_port[0]
        port = host_port[1] if host_port[1:] else None

        if port:
            port_num = int(port)

            # exclude things like '08090' or ' 8090'
            if port != str(port_num) or not (1 <= port_num < 65536):
                raise ValueError("Invalid port")

        return host, port
    except Exception:
        raise ValueError("Invalid server name '%s'" % server_name)


def is_valid_matrix_server_name(string: str) -> bool:
    """Validate that the given string is a valid Matrix server name.

    A string is a valid Matrix server name if it is one of the following, plus
    an optional port:

    a. IPv4 address
    b. IPv6 literal (`[IPV6_ADDRESS]`)
    c. A valid hostname

    :param string: The string to validate

    :return: Whether the input is a valid Matrix server name
    """

    try:
        host, port = parse_server_name(string)
    except ValueError:
        return False

    valid_ipv4_addr = isIPAddress(host)
    valid_ipv6_literal = (
        host[0] == "[" and host[-1] == "]" and isIPv6Address(host[1:-1])
    )

    return valid_ipv4_addr or valid_ipv6_literal or is_valid_hostname(host)


def normalise_address(address: str, medium: str) -> str:
    if medium == "email":
        return address.casefold()
    else:
        return address



# Copyright 2019 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import TYPE_CHECKING, Optional

from twisted.web.server import Request

from sydent.db.accounts import AccountStore
from sydent.http.servlets import MatrixRestError, get_args
from sydent.terms.terms import get_terms

if TYPE_CHECKING:
    from sydent.sydent import Sydent
    from sydent.users.accounts import Account

logger = logging.getLogger(__name__)


def tokenFromRequest(request: Request) -> Optional[str]:
    """Extract token from header of query parameter.

    :param request: The request to look for an access token in.

    :return: The token or None if not found
    """
    token = None
    # check for Authorization header first
    authHeader = request.getHeader("Authorization")
    if authHeader is not None and authHeader.startswith("Bearer "):
        token = authHeader[len("Bearer ") :]

    # no? try access_token query param
    if token is None:
        args = get_args(request, ("access_token",), required=False)
        token = args.get("access_token")

    return token


def authV2(
    sydent: "Sydent",
    request: Request,
    requireTermsAgreed: bool = True,
) -> "Account":
    """For v2 APIs check that the request has a valid access token associated with it

    :param sydent: The Sydent instance to use.
    :param request: The request to look for an access token in.
    :param requireTermsAgreed: Whether to deny authentication if the user hasn't accepted
        the terms of service.

    :returns Account: The account object if there is correct auth
    :raises MatrixRestError: If the request is v2 but could not be authed or the user has
        not accepted terms.
    """
    token = tokenFromRequest(request)

    if token is None:
        raise MatrixRestError(401, "M_UNAUTHORIZED", "Unauthorized")

    accountStore = AccountStore(sydent)

    account = accountStore.getAccountByToken(token)
    if account is None:
        raise MatrixRestError(401, "M_UNAUTHORIZED", "Unauthorized")

    if requireTermsAgreed:
        terms = get_terms(sydent)
        if (
            terms.getMasterVersion() is not None
            and account.consentVersion != terms.getMasterVersion()
        ):
            raise MatrixRestError(403, "M_TERMS_NOT_SIGNED", "Terms not signed")

    return account



# Copyright 2014-2016 OpenMarket Ltd
# Copyright 2019 New Vector Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import random
import time
from typing import Awaitable, Callable, Dict, List, SupportsInt, Tuple

import attr
from twisted.internet.error import ConnectError
from twisted.names import client, dns
from twisted.names.dns import Record_SRV, RRHeader
from twisted.names.error import DNSNameError, DomainError

logger = logging.getLogger(__name__)

SERVER_CACHE: Dict[bytes, List["Server"]] = {}


@attr.s(frozen=True, slots=True, auto_attribs=True)
class Server:
    """
    Our record of an individual server which can be tried to reach a destination.
    Attributes:
        host (bytes): target hostname
        port (int):
        priority (int):
        weight (int):
        expires (int): when the cache should expire this record - in *seconds* since
            the epoch
    """

    host: bytes
    port: int
    priority: int = 0
    weight: int = 0
    expires: int = 0


def pick_server_from_list(server_list: List[Server]) -> Tuple[bytes, int]:
    """Randomly choose a server from the server list.

    :param server_list: List of candidate servers.

    :returns a (host, port) pair for the chosen server.
    """
    if not server_list:
        raise RuntimeError("pick_server_from_list called with empty list")

    # TODO: currently we only use the lowest-priority servers. We should maintain a
    # cache of servers known to be "down" and filter them out

    min_priority = min(s.priority for s in server_list)
    eligible_servers = list(s for s in server_list if s.priority == min_priority)
    total_weight = sum(s.weight for s in eligible_servers)
    target_weight = random.randint(0, total_weight)

    for s in eligible_servers:
        target_weight -= s.weight

        if target_weight <= 0:
            return s.host, s.port

    # this should be impossible.
    raise RuntimeError(
        "pick_server_from_list got to end of eligible server list.",
    )


# The signature of twisted.names.client.lookupService, if you omit the timeout
# argument. This is unannotated, but we can deduce the signature as follows:
# 1. Return type is the return type of
#    twisted.internet.interfaces.IResolver.lookupService. Its type annotation
#    is incorrect; its docstring says that tuple entries are a _list_ of RRHeaders,
#    but the annotation says the entries are individual RRHeaders.
# 2. Because we're looking up SRV records, we know that the payload of the RRHeaders
#    will be Record_SRVs. I made RRHeader's stub generic over the type of its
#    payload to reflect this. But that's a lie compared to Twisted's actual
#    RRHeader Type, so we need to enclose these in strings.
LookupService = Callable[
    [str],
    Awaitable[
        Tuple[
            List["RRHeader[Record_SRV]"],
            List["RRHeader[object]"],
            List["RRHeader[object]"],
        ]
    ],
]


class SrvResolver:
    """Interface to the dns client to do SRV lookups, with result caching.
    The default resolver in twisted.names doesn't do any caching (it has a CacheResolver,
    but the cache never gets populated), so we add our own caching layer here.

    :param dns_client: Twisted resolver impl

    :param cache: cache object

    :param get_time: Clock implementation. Should return seconds since the epoch.
    """

    def __init__(
        self,
        lookup_service: LookupService = client.lookupService,
        cache: Dict[bytes, List[Server]] = SERVER_CACHE,
        get_time: Callable[[], SupportsInt] = time.time,
    ) -> None:
        self._lookup_service = lookup_service
        self._cache = cache
        self._get_time = get_time

    async def resolve_service(self, service_name: bytes) -> List["Server"]:
        """Look up a SRV record

        :param service_name: The record to look up.

        :returns a list of the SRV records, or an empty list if none found.
        """
        now = int(self._get_time())

        cache_entry = self._cache.get(service_name, None)
        if cache_entry:
            if all(s.expires > now for s in cache_entry):
                servers = list(cache_entry)
                return servers

        try:
            answers, _, _ = await self._lookup_service(service_name.decode())
        except DNSNameError:
            # TODO: cache this. We can get the SOA out of the exception, and use
            # the negative-TTL value.
            return []
        except DomainError as e:
            # We failed to resolve the name (other than a NameError)
            # Try something in the cache, else rereaise
            cache_entry = self._cache.get(service_name, None)
            if cache_entry:
                logger.warning(
                    "Failed to resolve %r, falling back to cache. %r", service_name, e
                )
                return list(cache_entry)
            else:
                raise e

        if (
            len(answers) == 1
            and answers[0].type == dns.SRV
            and answers[0].payload
            and answers[0].payload.target == dns.Name(b".")
        ):
            raise ConnectError("Service %s unavailable" % service_name.decode())

        servers = []

        for answer in answers:
            if answer.type != dns.SRV or not answer.payload:
                continue

            payload = answer.payload

            servers.append(
                Server(
                    host=payload.target.name,
                    port=payload.port,
                    priority=payload.priority,
                    weight=payload.weight,
                    expires=now + answer.ttl,
                )
            )

        self._cache[service_name] = list(servers)
        return servers



# Copyright 2019-2021 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import logging.handlers
import os
from configparser import DEFAULTSECT, ConfigParser
from typing import Dict

from sydent.config.crypto import CryptoConfig
from sydent.config.database import DatabaseConfig
from sydent.config.email import EmailConfig
from sydent.config.general import GeneralConfig
from sydent.config.http import HTTPConfig
from sydent.config.sms import SMSConfig

logger = logging.getLogger(__name__)

CONFIG_DEFAULTS = {
    "general": {
        "server.name": os.environ.get("SYDENT_SERVER_NAME", ""),
        "log.path": "",
        "log.level": "INFO",
        "pidfile.path": os.environ.get("SYDENT_PID_FILE", "sydent.pid"),
        "terms.path": "",
        "address_lookup_limit": "10000",  # Maximum amount of addresses in a single /lookup request
        # The root path to use for load templates. This should contain branded
        # directories. Each directory should contain the following templates:
        #
        # * invite_template.eml
        # * verification_template.eml
        # * verify_response_template.html
        "templates.path": "res",
        # The brand directory to use if no brand hint (or an invalid brand hint)
        # is provided by the request.
        "brand.default": "matrix-org",
        # The following can be added to your local config file to enable prometheus
        # support.
        # 'prometheus_port': '8080',  # The port to serve metrics on
        # 'prometheus_addr': '',  # The address to bind to. Empty string means bind to all.
        # The following can be added to your local config file to enable sentry support.
        # 'sentry_dsn': 'https://...'  # The DSN has configured in the sentry instance project.
        # Whether clients and homeservers can register an association using v1 endpoints. This
        # option is now deprecated and will be superceded by the option `enable_v1_access`
        "enable_v1_associations": "true",
        "delete_tokens_on_bind": "true",
        # Prevent outgoing requests from being sent to the following blacklisted
        # IP address CIDR ranges. If this option is not specified or empty then
        # it defaults to private IP address ranges.
        #
        # The blacklist applies to all outbound requests except replication
        # requests.
        #
        # (0.0.0.0 and :: are always blacklisted, whether or not they are
        # explicitly listed here, since they correspond to unroutable
        # addresses.)
        "ip.blacklist": "",
        # List of IP address CIDR ranges that should be allowed for outbound
        # requests. This is useful for specifying exceptions to wide-ranging
        # blacklisted target IP ranges.
        #
        # This whitelist overrides `ip.blacklist` and defaults to an empty
        # list.
        "ip.whitelist": "",
        # A list of homeservers that are allowed to register with this identity server. Defaults to
        # allowing all homeservers. If a list is specified, the config option `enable_v1_access` must be
        # set to 'false'.
        "homeserver_allow_list": "",
        # If set to 'false', entirely disable access via the V1 api.
        "enable_v1_access": "true",
    },
    "db": {
        "db.file": os.environ.get("SYDENT_DB_PATH", "sydent.db"),
    },
    "http": {
        "clientapi.http.bind_address": "::",
        "clientapi.http.port": "8090",
        "internalapi.http.bind_address": "::1",
        "internalapi.http.port": "",
        "replication.https.certfile": "",
        "replication.https.cacert": "",  # This should only be used for testing
        "replication.https.bind_address": "::",
        "replication.https.port": "4434",
        "obey_x_forwarded_for": "False",
        "federation.verifycerts": "True",
        # verify_response_template is deprecated, but still used if defined. Define
        # templates.path and brand.default under general instead.
        #
        # 'verify_response_template': 'res/verify_response_page_template',
        "client_http_base": "",
    },
    "email": {
        # email.template and email.invite_template are deprecated, but still used
        # if defined. Define templates.path and brand.default under general instead.
        #
        # 'email.template': 'res/verification_template.eml',
        # 'email.invite_template': 'res/invite_template.eml',
        "email.from": "Sydent Validation <noreply@{hostname}>",
        "email.subject": "Your Validation Token",
        "email.invite.subject": "%(sender_display_name)s has invited you to chat",
        "email.invite.subject_space": "%(sender_display_name)s has invited you to a space",
        "email.smtphost": "localhost",
        "email.smtpport": "25",
        "email.smtpusername": "",
        "email.smtppassword": "",
        "email.hostname": "",
        "email.tlsmode": "0",
        # The web client location which will be used if it is not provided by
        # the homeserver.
        #
        # This should be the scheme and hostname only, see res/invite_template.eml
        # for the full URL that gets generated.
        "email.default_web_client_location": "https://app.element.io",
        # When a user is invited to a room via their email address, that invite is
        # displayed in the room list using an obfuscated version of the user's email
        # address. These config options determine how much of the email address to
        # obfuscate. Note that the '@' sign is always included.
        #
        # If the string is longer than a configured limit below, it is truncated to that limit
        # with '...' added. Otherwise:
        #
        # * If the string is longer than 5 characters, it is truncated to 3 characters + '...'
        # * If the string is longer than 1 character, it is truncated to 1 character + '...'
        # * If the string is 1 character long, it is converted to '...'
        #
        # This ensures that a full email address is never shown, even if it is extremely
        # short.
        #
        # The number of characters from the beginning to reveal of the email's username
        # portion (left of the '@' sign)
        "email.third_party_invite_username_obfuscate_characters": "3",
        # The number of characters from the beginning to reveal of the email's domain
        # portion (right of the '@' sign)
        "email.third_party_invite_domain_obfuscate_characters": "3",
    },
    "sms": {
        "bodyTemplate": "Your code is {token}",
        "username": "",
        "password": "",
    },
    "crypto": {
        "ed25519.signingkey": "",
    },
}


class SydentConfig:
    """This is the class in charge of handling Sydent's configuration.
    Handling of each individual section is delegated to other classes
    stored in a `config_sections` list.

    To use this class, create a new object and then call one of
    `parse_config_file` or `parse_config_dict` before creating the
    Sydent object that uses it.
    """

    def __init__(self) -> None:
        self.general = GeneralConfig()
        self.database = DatabaseConfig()
        self.crypto = CryptoConfig()
        self.sms = SMSConfig()
        self.email = EmailConfig()
        self.http = HTTPConfig()

        self.config_sections = [
            self.general,
            self.database,
            self.crypto,
            self.sms,
            self.email,
            self.http,
        ]

    def _parse_config(self, cfg: ConfigParser) -> bool:
        """
        Run the parse_config method on each of the objects in
        self.config_sections

        :param cfg: the configuration to be parsed

        :return: whether or not cfg has been altered. This method CAN
            return True, but it *shouldn't* as this leads to altering the
            config file.
        """
        needs_saving = False
        for section in self.config_sections:
            if section.parse_config(cfg):
                needs_saving = True

        return needs_saving

    def parse_from_config_parser(self, cfg: ConfigParser) -> bool:
        """
        Parse the configuration from a ConfigParser object

        :param cfg: the configuration to be parsed

        :return: whether or not cfg has been altered. This method CAN
            return True, but it *shouldn't* as this leads to altering the
            config file.
        """
        return self._parse_config(cfg)

    def parse_config_file(self, config_file: str) -> None:
        """
        Parse the given config from a filepath, populating missing items and
        sections.

        :param config_file: the file to be parsed
        """
        # If the config file doesn't exist, prepopulate the config object
        # with the defaults, in the right section.
        #
        # Otherwise, we have to put the defaults in the DEFAULT section,
        # to ensure that they don't override anyone's settings which are
        # in their config file in the default section (which is likely,
        # because sydent used to be braindead).
        use_defaults = not os.path.exists(config_file)

        cfg = ConfigParser()
        for sect, entries in CONFIG_DEFAULTS.items():
            cfg.add_section(sect)
            for k, v in entries.items():
                cfg.set(DEFAULTSECT if use_defaults else sect, k, v)

        cfg.read(config_file)

        # TODO: Don't alter config file when starting Sydent so that
        #       it can be set to read-only
        needs_saving = self.parse_from_config_parser(cfg)

        if needs_saving:
            fp = open(config_file, "w")
            cfg.write(fp)
            fp.close()

    def parse_config_dict(self, config_dict: Dict[str, Dict[str, str]]) -> None:
        """
        Parse the given config from a dictionary, populating missing items and sections

        :param config_dict: the configuration dictionary to be parsed
        """
        # Build a config dictionary from the defaults merged with the given dictionary
        config = copy.deepcopy(CONFIG_DEFAULTS)
        for section, section_dict in config_dict.items():
            if section not in config:
                config[section] = {}
            for option in section_dict.keys():
                config[section][option] = config_dict[section][option]

        # Build a ConfigParser from the merged dictionary
        cfg = ConfigParser()
        for section, section_dict in config.items():
            cfg.add_section(section)
            for option, value in section_dict.items():
                cfg.set(section, option, value)

        self.parse_from_config_parser(cfg)



# Copyright 2014 OpenMarket Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from io import BytesIO
from typing import TYPE_CHECKING, Optional, cast

import twisted.internet.ssl
from twisted.internet import defer, protocol
from twisted.internet._sslverify import IOpenSSLTrustRoot
from twisted.internet.interfaces import ITCPTransport
from twisted.internet.protocol import connectionDone
from twisted.python.failure import Failure
from twisted.web import server
from twisted.web.client import ResponseDone
from twisted.web.http import PotentialDataLoss
from twisted.web.iweb import UNKNOWN_LENGTH, IResponse

if TYPE_CHECKING:
    from sydent.sydent import Sydent


logger = logging.getLogger(__name__)

# Arbitrarily limited to 512 KiB.
MAX_REQUEST_SIZE = 512 * 1024


class SslComponents:
    def __init__(self, sydent: "Sydent") -> None:
        self.sydent = sydent

        self.myPrivateCertificate = self.makeMyCertificate()
        self.trustRoot = self.makeTrustRoot()

    def makeMyCertificate(self) -> Optional[twisted.internet.ssl.PrivateCertificate]:
        # TODO Move some of this loading into parse_config
        privKeyAndCertFilename = self.sydent.config.http.cert_file

        if privKeyAndCertFilename == "":
            logger.warning(
                "No HTTPS private key / cert found: not starting replication server "
                "or doing replication pushes"
            )
            return None

        try:
            fp = open(privKeyAndCertFilename)
        except OSError:
            logger.warning(
                "Unable to read private key / cert file from %s: not starting the replication HTTPS server "
                "or doing replication pushes.",
                privKeyAndCertFilename,
            )
            return None

        authData = fp.read()
        fp.close()
        return twisted.internet.ssl.PrivateCertificate.loadPEM(authData)

    def makeTrustRoot(self) -> IOpenSSLTrustRoot:
        # If this option is specified, use a specific root CA cert. This is useful for testing when it's not
        # practical to get the client cert signed by a real root CA but should never be used on a production server.
        caCertFilename = self.sydent.config.http.ca_cert_file
        if len(caCertFilename) > 0:
            try:
                fp = open(caCertFilename)
                caCert = twisted.internet.ssl.Certificate.loadPEM(fp.read())
                fp.close()
            except Exception:
                logger.warning("Failed to open CA cert file %s", caCertFilename)
                raise
            logger.warning("Using custom CA cert file: %s", caCertFilename)
            # Type ignore: I'm not going to add a stub for the semiprivate
            # _sslverify module. I've already taken on too much stubbing as it is!
            return twisted.internet._sslverify.OpenSSLCertificateAuthorities(  # type: ignore
                [caCert.original]
            )
        else:
            return twisted.internet.ssl.OpenSSLDefaultPaths()


class BodyExceededMaxSize(Exception):
    """The maximum allowed size of the HTTP body was exceeded."""


class _DiscardBodyWithMaxSizeProtocol(protocol.Protocol):
    """A protocol which immediately errors upon receiving data."""

    transport: ITCPTransport

    def __init__(self, deferred: "defer.Deferred[bytes]") -> None:
        self.deferred = deferred

    def _maybe_fail(self) -> None:
        """
        Report a max size exceed error and disconnect the first time this is called.
        """
        if not self.deferred.called:
            self.deferred.errback(BodyExceededMaxSize())
            # Close the connection (forcefully) since all the data will get
            # discarded anyway.
            self.transport.abortConnection()

    def dataReceived(self, data: bytes) -> None:
        self._maybe_fail()

    def connectionLost(self, reason: Failure = connectionDone) -> None:
        self._maybe_fail()


class _ReadBodyWithMaxSizeProtocol(protocol.Protocol):
    """A protocol which reads body to a stream, erroring if the body exceeds a maximum size."""

    transport: ITCPTransport

    def __init__(
        self, deferred: "defer.Deferred[bytes]", max_size: Optional[int]
    ) -> None:
        self.stream = BytesIO()
        self.deferred = deferred
        self.length = 0
        self.max_size = max_size

    def dataReceived(self, data: bytes) -> None:
        # If the deferred was called, bail early.
        if self.deferred.called:
            return

        self.stream.write(data)
        self.length += len(data)
        # The first time the maximum size is exceeded, error and cancel the
        # connection. dataReceived might be called again if data was received
        # in the meantime.
        if self.max_size is not None and self.length >= self.max_size:
            self.deferred.errback(BodyExceededMaxSize())
            # Close the connection (forcefully) since all the data will get
            # discarded anyway.
            if self.transport is not None:
                self.transport.abortConnection()

    def connectionLost(self, reason: Failure = connectionDone) -> None:
        # If the maximum size was already exceeded, there's nothing to do.
        if self.deferred.called:
            return

        if reason.check(ResponseDone):
            self.deferred.callback(self.stream.getvalue())
        elif reason.check(PotentialDataLoss):
            # stolen from https://github.com/twisted/treq/pull/49/files
            # http://twistedmatrix.com/trac/ticket/4840
            self.deferred.callback(self.stream.getvalue())
        else:
            self.deferred.errback(reason)


def read_body_with_max_size(
    response: IResponse, max_size: Optional[int]
) -> "defer.Deferred[bytes]":
    """
    Read a HTTP response body to a file-object. Optionally enforcing a maximum file size.

    If the maximum file size is reached, the returned Deferred will resolve to a
    Failure with a BodyExceededMaxSize exception.

    Args:
        response: The HTTP response to read from.
        max_size: The maximum file size to allow.

    Returns:
        A Deferred which resolves to the read body.
    """
    d: "defer.Deferred[bytes]" = defer.Deferred()

    # If the Content-Length header gives a size larger than the maximum allowed
    # size, do not bother downloading the body.
    # Type safety: twisted guarantees that response.length is either the
    # "opaque" object UNKNOWN_LENGTH, or else an int.
    if max_size is not None and response.length != UNKNOWN_LENGTH:
        response.length = cast(int, response.length)
        if response.length > max_size:
            response.deliverBody(_DiscardBodyWithMaxSizeProtocol(d))
            return d

    response.deliverBody(_ReadBodyWithMaxSizeProtocol(d, max_size))
    return d


class SizeLimitingRequest(server.Request):
    def handleContentChunk(self, data: bytes) -> None:
        if self.content.tell() + len(data) > MAX_REQUEST_SIZE:
            logger.info(
                "Aborting connection from %s because the request exceeds maximum size",
                # Formerly `self.client.host`, but `host` isn't provided by `IAddress`
                self.client,
            )
            assert self.transport is not None
            self.transport.abortConnection()
            return

        return super().handleContentChunk(data)



#  Copyright 2021 The Matrix.org Foundation C.I.C.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from unittest.mock import patch

from twisted.internet.error import DNSLookupError
from twisted.test.proto_helpers import StringTransport
from twisted.trial.unittest import TestCase
from twisted.web.client import Agent

from sydent.http.blacklisting_reactor import BlacklistingReactorWrapper
from sydent.http.srvresolver import Server
from tests.utils import AsyncMock, make_request, make_sydent


class BlacklistingAgentTest(TestCase):
    def setUp(self):
        config = {
            "general": {
                "ip.blacklist": "5.0.0.0/8",
                "ip.whitelist": "5.1.1.1",
            },
        }

        self.sydent = make_sydent(test_config=config)

        self.reactor = self.sydent.reactor

        self.safe_domain, self.safe_ip = b"safe.test", b"1.2.3.4"
        self.unsafe_domain, self.unsafe_ip = b"danger.test", b"5.6.7.8"
        self.allowed_domain, self.allowed_ip = b"allowed.test", b"5.1.1.1"

        # Configure the reactor's DNS resolver.
        for (domain, ip) in (
            (self.safe_domain, self.safe_ip),
            (self.unsafe_domain, self.unsafe_ip),
            (self.allowed_domain, self.allowed_ip),
        ):
            self.reactor.lookups[domain.decode()] = ip.decode()
            self.reactor.lookups[ip.decode()] = ip.decode()

        self.ip_whitelist = self.sydent.config.general.ip_whitelist
        self.ip_blacklist = self.sydent.config.general.ip_blacklist

    def test_reactor(self):
        """Apply the blacklisting reactor and ensure it properly blocks
        connections to particular domains and IPs.
        """
        agent = Agent(
            BlacklistingReactorWrapper(
                self.reactor,
                ip_whitelist=self.ip_whitelist,
                ip_blacklist=self.ip_blacklist,
            ),
        )

        # The unsafe domains and IPs should be rejected.
        for domain in (self.unsafe_domain, self.unsafe_ip):
            self.failureResultOf(
                agent.request(b"GET", b"http://" + domain), DNSLookupError
            )

        self.reactor.tcpClients = []

        # The safe domains IPs should be accepted.
        for domain in (
            self.safe_domain,
            self.allowed_domain,
            self.safe_ip,
            self.allowed_ip,
        ):
            agent.request(b"GET", b"http://" + domain)

            # Grab the latest TCP connection.
            (
                host,
                port,
                client_factory,
                _timeout,
                _bindAddress,
            ) = self.reactor.tcpClients.pop()

    @patch(
        "sydent.http.srvresolver.SrvResolver.resolve_service", new_callable=AsyncMock
    )
    def test_federation_client_allowed_ip(self, resolver):
        self.sydent.run()

        resolver.return_value = [
            Server(
                host=self.allowed_domain,
                port=443,
                priority=1,
                weight=1,
                expires=100,
            )
        ]

        request, channel = make_request(
            self.sydent.reactor,
            self.sydent.clientApiHttpServer.factory,
            "POST",
            "/_matrix/identity/v2/account/register",
            {
                "access_token": "foo",
                "expires_in": 300,
                "matrix_server_name": "example.com",
                "token_type": "Bearer",
            },
        )

        transport, protocol = self._get_http_request(
            self.allowed_ip.decode("ascii"), 443
        )

        self.assertRegex(
            transport.value(), b"^GET /_matrix/federation/v1/openid/userinfo"
        )
        self.assertRegex(transport.value(), b"Host: example.com")

        # Send it the HTTP response
        res_json = b'{ "sub": "@test:example.com" }'
        protocol.dataReceived(
            b"HTTP/1.1 200 OK\r\n"
            b"Server: Fake\r\n"
            b"Content-Type: application/json\r\n"
            b"Content-Length: %i\r\n"
            b"\r\n"
            b"%s" % (len(res_json), res_json)
        )

        self.assertEqual(channel.code, 200)

    @patch(
        "sydent.http.srvresolver.SrvResolver.resolve_service", new_callable=AsyncMock
    )
    def test_federation_client_safe_ip(self, resolver):
        self.sydent.run()

        resolver.return_value = [
            Server(
                host=self.safe_domain,
                port=443,
                priority=1,
                weight=1,
                expires=100,
            )
        ]

        request, channel = make_request(
            self.sydent.reactor,
            self.sydent.clientApiHttpServer.factory,
            "POST",
            "/_matrix/identity/v2/account/register",
            {
                "access_token": "foo",
                "expires_in": 300,
                "matrix_server_name": "example.com",
                "token_type": "Bearer",
            },
        )

        transport, protocol = self._get_http_request(self.safe_ip.decode("ascii"), 443)

        self.assertRegex(
            transport.value(), b"^GET /_matrix/federation/v1/openid/userinfo"
        )
        self.assertRegex(transport.value(), b"Host: example.com")

        # Send it the HTTP response
        res_json = b'{ "sub": "@test:example.com" }'
        protocol.dataReceived(
            b"HTTP/1.1 200 OK\r\n"
            b"Server: Fake\r\n"
            b"Content-Type: application/json\r\n"
            b"Content-Length: %i\r\n"
            b"\r\n"
            b"%s" % (len(res_json), res_json)
        )

        self.assertEqual(channel.code, 200)

    @patch("sydent.http.srvresolver.SrvResolver.resolve_service")
    def test_federation_client_unsafe_ip(self, resolver):
        self.sydent.run()

        resolver.return_value = [
            Server(
                host=self.unsafe_domain,
                port=443,
                priority=1,
                weight=1,
                expires=100,
            )
        ]

        request, channel = make_request(
            self.sydent.reactor,
            self.sydent.clientApiHttpServer.factory,
            "POST",
            "/_matrix/identity/v2/account/register",
            {
                "access_token": "foo",
                "expires_in": 300,
                "matrix_server_name": "example.com",
                "token_type": "Bearer",
            },
        )

        self.assertNot(self.reactor.tcpClients)

        self.assertEqual(channel.code, 500)

    def _get_http_request(self, expected_host, expected_port):
        clients = self.reactor.tcpClients
        (host, port, factory, _timeout, _bindAddress) = clients[-1]
        self.assertEqual(host, expected_host)
        self.assertEqual(port, expected_port)

        # complete the connection and wire it up to a fake transport
        protocol = factory.buildProtocol(None)
        transport = StringTransport()
        protocol.makeConnection(transport)

        return transport, protocol



#  Copyright 2021 The Matrix.org Foundation C.I.C.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Optional
from unittest.mock import Mock, patch

from twisted.trial import unittest

from sydent.types import JsonDict
from tests.utils import make_request, make_sydent


class TestRequestCode(unittest.TestCase):
    def setUp(self) -> None:
        # Create a new sydent
        self.sydent = make_sydent()

    def _make_request(self, url: str, body: Optional[JsonDict] = None) -> Mock:
        # Patch out the email sending so we can investigate the resulting email.
        with patch("sydent.util.emailutils.smtplib") as smtplib:
            request, channel = make_request(
                self.sydent.reactor,
                self.sydent.clientApiHttpServer.factory,
                "POST",
                url,
                body,
            )

        self.assertEqual(channel.code, 200)

        # Fish out the SMTP object and return it.
        smtp = smtplib.SMTP.return_value
        smtp.sendmail.assert_called_once()

        return smtp

    def test_request_code(self) -> None:
        self.sydent.run()

        smtp = self._make_request(
            "/_matrix/identity/api/v1/validate/email/requestToken",
            {
                "email": "test@test",
                "client_secret": "oursecret",
                "send_attempt": 0,
            },
        )

        # Ensure the email is as expected.
        email_contents = smtp.sendmail.call_args[0][2].decode("utf-8")
        self.assertIn("Confirm your email address for Matrix", email_contents)

    def test_request_code_via_url_query_params(self) -> None:
        self.sydent.run()
        url = (
            "/_matrix/identity/api/v1/validate/email/requestToken?"
            "email=test@test"
            "&client_secret=oursecret"
            "&send_attempt=0"
        )
        smtp = self._make_request(url)

        # Ensure the email is as expected.
        email_contents = smtp.sendmail.call_args[0][2].decode("utf-8")
        self.assertIn("Confirm your email address for Matrix", email_contents)

    def test_branded_request_code(self) -> None:
        self.sydent.run()

        smtp = self._make_request(
            "/_matrix/identity/api/v1/validate/email/requestToken?brand=vector-im",
            {
                "email": "test@test",
                "client_secret": "oursecret",
                "send_attempt": 0,
            },
        )

        # Ensure the email is as expected.
        email_contents = smtp.sendmail.call_args[0][2].decode("utf-8")
        self.assertIn("Confirm your email address for Element", email_contents)



# Copyright 2021 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path
import urllib
from unittest.mock import Mock, patch

from twisted.trial import unittest

from sydent.util.emailutils import sendEmail
from tests.utils import make_sydent


class TestTemplate(unittest.TestCase):
    def setUp(self):
        # Create a new sydent
        self.sydent = make_sydent()

    def test_jinja_vector_invite(self):
        substitutions = {
            "address": "foo@example.com",
            "medium": "email",
            "room_alias": "#somewhere:exmaple.org",
            "room_avatar_url": "mxc://example.org/s0meM3dia",
            "room_id": "!something:example.org",
            "room_name": "Bob's Emporium of Messages",
            "sender": "@bob:example.com",
            "sender_avatar_url": "mxc://example.org/an0th3rM3dia",
            "sender_display_name": "<Bob Smith>",
            "bracketed_verified_sender": "Bob Smith",
            "bracketed_room_name": "Bob's Emporium of Messages",
            "to": "person@test.test",
            "token": "a_token",
            "ephemeral_private_key": "mystery_key",
            "web_client_location": "https://app.element.io",
            "room_type": "",
        }

        # self.sydent.config.email.invite_template is deprecated
        if self.sydent.config.email.invite_template is None:
            templateFile = self.sydent.get_branded_template(
                "vector-im",
                "invite_template.eml",
            )
        else:
            templateFile = self.sydent.config.email.invite_template

        with patch("sydent.util.emailutils.smtplib") as smtplib:
            sendEmail(self.sydent, templateFile, "test@test.com", substitutions)

        smtp = smtplib.SMTP.return_value
        email_contents = smtp.sendmail.call_args[0][2].decode("utf-8")

        # test url input is encoded
        self.assertIn(urllib.parse.quote("mxc://example.org/s0meM3dia"), email_contents)

        # test html input is escaped
        self.assertIn("Bob&#39;s Emporium of Messages", email_contents)

        # test safe values are not escaped
        self.assertIn("<Bob Smith>", email_contents)

        # test our link is as expected
        expected_url = (
            "https://app.element.io/#/room/"
            + urllib.parse.quote("!something:example.org")
            + "?email="
            + urllib.parse.quote("test@test.com")
            + "&signurl=https%3A%2F%2Fvector.im%2F_matrix%2Fidentity%2Fapi%2Fv1%2Fsign-ed25519%3Ftoken%3D"
            + urllib.parse.quote("a_token")
            + "%26private_key%3D"
            + urllib.parse.quote("mystery_key")
            + "&room_name="
            + urllib.parse.quote("Bob's Emporium of Messages")
            + "&room_avatar_url="
            + urllib.parse.quote("mxc://example.org/s0meM3dia")
            + "&inviter_name="
            + urllib.parse.quote("<Bob Smith>")
            + "&guest_access_token=&guest_user_id=&room_type="
        )
        text = email_contents.splitlines()
        link = text[19]
        self.assertEqual(link, expected_url)

    def test_jinja_matrix_invite(self):
        substitutions = {
            "address": "foo@example.com",
            "medium": "email",
            "room_alias": "#somewhere:exmaple.org",
            "room_avatar_url": "mxc://example.org/s0meM3dia",
            "room_id": "!something:example.org",
            "room_name": "Bob's Emporium of Messages",
            "sender": "@bob:example.com",
            "sender_avatar_url": "mxc://example.org/an0th3rM3dia",
            "sender_display_name": "<Bob Smith>",
            "bracketed_verified_sender": "Bob Smith",
            "bracketed_room_name": "Bob's Emporium of Messages",
            "to": "person@test.test",
            "token": "a_token",
            "ephemeral_private_key": "mystery_key",
            "web_client_location": "https://matrix.org",
            "room_type": "",
        }

        # self.sydent.config.email.invite_template is deprecated
        if self.sydent.config.email.invite_template is None:
            templateFile = self.sydent.get_branded_template(
                "matrix-org",
                "invite_template.eml",
            )
        else:
            templateFile = self.sydent.config.email.invite_template

        with patch("sydent.util.emailutils.smtplib") as smtplib:
            sendEmail(self.sydent, templateFile, "test@test.com", substitutions)

        smtp = smtplib.SMTP.return_value
        email_contents = smtp.sendmail.call_args[0][2].decode("utf-8")

        # test url input is encoded
        self.assertIn(urllib.parse.quote("mxc://example.org/s0meM3dia"), email_contents)

        # test html input is escaped
        self.assertIn("Bob&#39;s Emporium of Messages", email_contents)

        # test safe values are not escaped
        self.assertIn("<Bob Smith>", email_contents)

        # test our link is as expected
        expected_url = (
            "https://matrix.org/#/room/"
            + urllib.parse.quote("!something:example.org")
            + "?email="
            + urllib.parse.quote("test@test.com")
            + "&signurl=https%3A%2F%2Fmatrix.org%2F_matrix%2Fidentity%2Fapi%2Fv1%2Fsign-ed25519%3Ftoken%3D"
            + urllib.parse.quote("a_token")
            + "%26private_key%3D"
            + urllib.parse.quote("mystery_key")
            + "&room_name="
            + urllib.parse.quote("Bob's Emporium of Messages")
            + "&room_avatar_url="
            + urllib.parse.quote("mxc://example.org/s0meM3dia")
            + "&inviter_name="
            + urllib.parse.quote("<Bob Smith>")
            + "&guest_access_token=&guest_user_id=&room_type="
        )
        text = email_contents.splitlines()
        link = text[22]
        self.assertEqual(link, expected_url)

    def test_jinja_matrix_verification(self):
        substitutions = {
            "address": "foo@example.com",
            "medium": "email",
            "to": "person@test.test",
            "token": "<<token>>",
            "link": "https://link_test.com",
        }

        templateFile = self.sydent.get_branded_template(
            "matrix-org",
            "verification_template.eml",
        )

        with patch("sydent.util.emailutils.smtplib") as smtplib:
            sendEmail(self.sydent, templateFile, "test@test.com", substitutions)

        smtp = smtplib.SMTP.return_value
        email_contents = smtp.sendmail.call_args[0][2].decode("utf-8")

        # test html input is escaped
        self.assertIn("&lt;&lt;token&gt;&gt;", email_contents)

        # test safe values are not escaped
        self.assertIn("<<token>>", email_contents)

    @patch(
        "sydent.util.emailutils.generateAlphanumericTokenOfLength",
        Mock(return_value="aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"),
    )
    def test_jinja_vector_verification(self):
        substitutions = {
            "address": "foo@example.com",
            "medium": "email",
            "to": "person@test.test",
            "link": "https://link_test.com",
        }

        templateFile = self.sydent.get_branded_template(
            "vector-im",
            "verification_template.eml",
        )

        with patch("sydent.util.emailutils.smtplib") as smtplib:
            sendEmail(self.sydent, templateFile, "test@test.com", substitutions)

        smtp = smtplib.SMTP.return_value
        email_contents = smtp.sendmail.call_args[0][2].decode("utf-8")

        path = os.path.join(
            self.sydent.config.general.templates_path,
            "vector_verification_sample.txt",
        )

        with open(path, "r") as file:
            expected_text = file.read()

        # remove the email headers as they are variable
        email_contents = email_contents[email_contents.index("Hello") :]

        # test all ouput is as expected
        self.assertEqual(email_contents, expected_text)



# Copyright 2021 Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
from unittest.mock import patch

from twisted.trial import unittest

from scripts.casefold_db import (
    calculate_lookup_hash,
    update_global_associations,
    update_local_associations,
)
from sydent.util import json_decoder
from sydent.util.emailutils import sendEmail
from tests.utils import make_sydent


class MigrationTestCase(unittest.TestCase):
    def create_signedassoc(self, medium, address, mxid, ts, not_before, not_after):
        return {
            "medium": medium,
            "address": address,
            "mxid": mxid,
            "ts": ts,
            "not_before": not_before,
            "not_after": not_after,
        }

    def setUp(self):
        # Create a new sydent
        self.sydent = make_sydent()

        # create some local associations
        associations = []

        for i in range(10):
            address = "bob%d@example.com" % i
            associations.append(
                {
                    "medium": "email",
                    "address": address,
                    "lookup_hash": calculate_lookup_hash(self.sydent, address),
                    "mxid": "@bob%d:example.com" % i,
                    "ts": (i * 10000),
                    "not_before": 0,
                    "not_after": 99999999999,
                }
            )
        # create some casefold-conflicting associations
        for i in range(5):
            address = "BOB%d@example.com" % i
            associations.append(
                {
                    "medium": "email",
                    "address": address,
                    "lookup_hash": calculate_lookup_hash(self.sydent, address),
                    "mxid": "@otherbob%d:example.com" % i,
                    "ts": (i * 10000),
                    "not_before": 0,
                    "not_after": 99999999999,
                }
            )

        associations.append(
            {
                "medium": "email",
                "address": "BoB4@example.com",
                "lookup_hash": calculate_lookup_hash(self.sydent, "BoB4@example.com"),
                "mxid": "@otherbob4:example.com",
                "ts": 42000,
                "not_before": 0,
                "not_after": 99999999999,
            }
        )

        # add all associations to db
        cur = self.sydent.db.cursor()

        cur.executemany(
            "INSERT INTO  local_threepid_associations "
            "(medium, address, lookup_hash, mxid, ts, notBefore, notAfter) "
            "VALUES (?, ?, ?, ?, ?, ?, ?)",
            [
                (
                    assoc["medium"],
                    assoc["address"],
                    assoc["lookup_hash"],
                    assoc["mxid"],
                    assoc["ts"],
                    assoc["not_before"],
                    assoc["not_after"],
                )
                for assoc in associations
            ],
        )

        self.sydent.db.commit()

        # create some global associations
        associations = []
        originServer = self.sydent.config.general.server_name

        for i in range(10):
            address = "bob%d@example.com" % i
            mxid = "@bob%d:example.com" % i
            ts = 10000 * i
            associations.append(
                {
                    "medium": "email",
                    "address": address,
                    "lookup_hash": calculate_lookup_hash(self.sydent, address),
                    "mxid": mxid,
                    "ts": ts,
                    "not_before": 0,
                    "not_after": 99999999999,
                    "originServer": originServer,
                    "originId": i,
                    "sgAssoc": json.dumps(
                        self.create_signedassoc(
                            "email", address, mxid, ts, 0, 99999999999
                        )
                    ),
                }
            )
        # create some casefold-conflicting associations
        for i in range(5):
            address = "BOB%d@example.com" % i
            mxid = "@BOB%d:example.com" % i
            ts = 10000 * i
            associations.append(
                {
                    "medium": "email",
                    "address": address,
                    "lookup_hash": calculate_lookup_hash(self.sydent, address),
                    "mxid": mxid,
                    "ts": ts + 1,
                    "not_before": 0,
                    "not_after": 99999999999,
                    "originServer": originServer,
                    "originId": i + 10,
                    "sgAssoc": json.dumps(
                        self.create_signedassoc(
                            "email", address, mxid, ts, 0, 99999999999
                        )
                    ),
                }
            )

        # add all associations to db
        cur = self.sydent.db.cursor()

        cur.executemany(
            "INSERT INTO global_threepid_associations "
            "(medium, address, lookup_hash, mxid, ts, notBefore, notAfter, originServer, originId, sgAssoc) "
            "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
            [
                (
                    assoc["medium"],
                    assoc["address"],
                    assoc["lookup_hash"],
                    assoc["mxid"],
                    assoc["ts"],
                    assoc["not_before"],
                    assoc["not_after"],
                    assoc["originServer"],
                    assoc["originId"],
                    assoc["sgAssoc"],
                )
                for assoc in associations
            ],
        )

        self.sydent.db.commit()

    def test_migration_email(self):
        with patch("sydent.util.emailutils.smtplib") as smtplib:
            # self.sydent.config.email.template is deprecated
            if self.sydent.config.email.template is None:
                templateFile = self.sydent.get_branded_template(
                    None,
                    "migration_template.eml",
                )
            else:
                templateFile = self.sydent.config.email.template

            sendEmail(
                self.sydent,
                templateFile,
                "bob@example.com",
                {
                    "mxid": "@bob:example.com",
                    "subject_header_value": "MatrixID Deletion",
                },
            )
            smtp = smtplib.SMTP.return_value
            email_contents = smtp.sendmail.call_args[0][2].decode("utf-8")
            self.assertIn("In the past", email_contents)

            # test email was sent
            smtp.sendmail.assert_called()

    def test_local_db_migration(self):
        with patch("sydent.util.emailutils.smtplib") as smtplib:
            update_local_associations(
                self.sydent,
                self.sydent.db,
                send_email=True,
                dry_run=False,
                test=True,
            )

        # test 5 emails were sent
        smtp = smtplib.SMTP.return_value
        self.assertEqual(smtp.sendmail.call_count, 5)

        # don't send emails to people who weren't affected
        self.assertNotIn(
            smtp.sendmail.call_args_list,
            [
                "bob5@example.com",
                "bob6@example.com",
                "bob7@example.com",
                "bob8@example.com",
                "bob9@example.com",
            ],
        )

        # make sure someone who is affected gets email
        self.assertIn("bob4@example.com", smtp.sendmail.call_args_list[0][0])

        cur = self.sydent.db.cursor()
        res = cur.execute("SELECT * FROM local_threepid_associations")

        db_state = res.fetchall()

        # five addresses should have been deleted
        self.assertEqual(len(db_state), 10)

        # iterate through db and make sure all addresses are casefolded and hash matches casefolded address
        for row in db_state:
            casefolded = row[2].casefold()
            self.assertEqual(row[2], casefolded)
            self.assertEqual(
                calculate_lookup_hash(self.sydent, row[2]),
                calculate_lookup_hash(self.sydent, casefolded),
            )

    def test_global_db_migration(self):
        update_global_associations(
            self.sydent,
            self.sydent.db,
            dry_run=False,
        )

        cur = self.sydent.db.cursor()
        res = cur.execute("SELECT * FROM global_threepid_associations")

        db_state = res.fetchall()

        # five addresses should have been deleted
        self.assertEqual(len(db_state), 10)

        # iterate through db and make sure all addresses are casefolded and hash matches casefolded address
        # and make sure the casefolded address matches the address in sgAssoc
        for row in db_state:
            casefolded = row[2].casefold()
            self.assertEqual(row[2], casefolded)
            self.assertEqual(
                calculate_lookup_hash(self.sydent, row[2]),
                calculate_lookup_hash(self.sydent, casefolded),
            )
            sgassoc = json_decoder.decode(row[9])
            self.assertEqual(row[2], sgassoc["address"])

    def test_local_no_email_does_not_send_email(self):
        with patch("sydent.util.emailutils.smtplib") as smtplib:
            update_local_associations(
                self.sydent,
                self.sydent.db,
                send_email=False,
                dry_run=False,
                test=True,
            )
            smtp = smtplib.SMTP.return_value

            # test no emails were sent
            self.assertEqual(smtp.sendmail.call_count, 0)

    def test_dry_run_does_nothing(self):
        # reset DB
        self.setUp()

        cur = self.sydent.db.cursor()

        # grab a snapshot of global table before running script
        res1 = cur.execute("SELECT mxid FROM global_threepid_associations")
        list1 = res1.fetchall()

        with patch("sydent.util.emailutils.smtplib") as smtplib:
            update_global_associations(
                self.sydent,
                self.sydent.db,
                dry_run=True,
            )

        # test no emails were sent
        smtp = smtplib.SMTP.return_value
        self.assertEqual(smtp.sendmail.call_count, 0)

        res2 = cur.execute("SELECT mxid FROM global_threepid_associations")
        list2 = res2.fetchall()

        self.assertEqual(list1, list2)

        # grab a snapshot of local table db before running script
        res3 = cur.execute("SELECT mxid FROM local_threepid_associations")
        list3 = res3.fetchall()

        with patch("sydent.util.emailutils.smtplib") as smtplib:
            update_local_associations(
                self.sydent,
                self.sydent.db,
                send_email=True,
                dry_run=True,
                test=True,
            )

        # test no emails were sent
        smtp = smtplib.SMTP.return_value
        self.assertEqual(smtp.sendmail.call_count, 0)

        res4 = cur.execute("SELECT mxid FROM local_threepid_associations")
        list4 = res4.fetchall()
        self.assertEqual(list3, list4)



# See "Request body" section of
# https://www.openmarket.com/docs/Content/apis/v4http/send-json.htm
from typing_extensions import Literal, TypedDict

TypeOfNumber = Literal[1, 3, 5]


class _MessageRequired(TypedDict):
    type: Literal["text", "hexEncodedText", "binary", "wapPush"]
    content: str


class Message(_MessageRequired, total=False):
    charset: Literal["GSM", "Latin-1", "UTF-8", "UTF-16"]
    validityPeriod: int
    url: str
    mlc: Literal["reject", "truncate", "segment"]
    udh: bool


class _DestinationRequired(TypedDict):
    address: str


class Destination(_DestinationRequired, total=False):
    mobileOperatorId: int


class _SourceRequired(TypedDict):
    address: str


class Source(_SourceRequired, total=False):
    ton: TypeOfNumber


class _MobileTerminateRequired(TypedDict):
    # OpenMarket says these are required fields
    destination: Destination
    message: Message


class MobileTerminate(_MobileTerminateRequired, total=False):
    # And these are all optional.
    interaction: Literal["one-way", "two-way"]
    promotional: bool  # Ignored, unless we're sending to India
    source: Source
    # The API also offers optional "options" and "delivery" keys,
    # which we don't use


class SendSMSBody(TypedDict):
    mobileTerminate: MobileTerminate



# Copyright 2021 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from http import HTTPStatus
from unittest.mock import patch

import twisted.internet.error
import twisted.web.client
from parameterized import parameterized
from twisted.trial import unittest

from tests.utils import make_request, make_sydent


class ThreepidUnbindTestCase(unittest.TestCase):
    """Tests Sydent's threepidunbind servlet"""

    def setUp(self) -> None:
        # Create a new sydent
        self.sydent = make_sydent()

    # Duplicated from TestRegisterServelet. Is there a way for us to keep
    # ourselves DRY?
    @parameterized.expand(
        [
            (twisted.internet.error.DNSLookupError(),),
            (twisted.internet.error.TimeoutError(),),
            (twisted.internet.error.ConnectionRefusedError(),),
            # Naughty: strictly we're supposed to initialise a ResponseNeverReceived
            # with a list of 1 or more failures.
            (twisted.web.client.ResponseNeverReceived([]),),
        ]
    )
    def test_connection_failure(self, exc: Exception) -> None:
        """Check we respond sensibly if we can't contact the homeserver."""
        self.sydent.run()
        with patch.object(
            self.sydent.sig_verifier, "authenticate_request", side_effect=exc
        ):
            request, channel = make_request(
                self.sydent.reactor,
                self.sydent.clientApiHttpServer.factory,
                "POST",
                "/_matrix/identity/v2/3pid/unbind",
                content={
                    "mxid": "@alice:wonderland",
                    "threepid": {
                        "address": "alice.cooper@wonderland.biz",
                        "medium": "email",
                    },
                },
            )
        self.assertEqual(channel.code, HTTPStatus.INTERNAL_SERVER_ERROR)
        self.assertEqual(channel.json_body["errcode"], "M_UNKNOWN")
        self.assertIn("contact", channel.json_body["error"])



# Copyright 2014 OpenMarket Ltd
# Copyright 2019 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

from twisted.web.server import Request

from sydent.db.valsession import ThreePidValSessionStore
from sydent.http.auth import authV2
from sydent.http.servlets import (
    MatrixRestError,
    SydentResource,
    get_args,
    jsonwrap,
    send_cors,
)
from sydent.types import JsonDict
from sydent.util.stringutils import is_valid_client_secret
from sydent.validators import (
    IncorrectClientSecretException,
    InvalidSessionIdException,
    SessionExpiredException,
    SessionNotValidatedException,
)

if TYPE_CHECKING:
    from sydent.sydent import Sydent


class ThreePidBindServlet(SydentResource):
    def __init__(self, sydent: "Sydent", require_auth: bool = False) -> None:
        super().__init__()
        self.sydent = sydent
        self.require_auth = require_auth

    @jsonwrap
    def render_POST(self, request: Request) -> JsonDict:
        send_cors(request)

        account = None
        if self.require_auth:
            account = authV2(self.sydent, request)

        args = get_args(request, ("sid", "client_secret", "mxid"))

        sid = args["sid"]
        mxid = args["mxid"]
        clientSecret = args["client_secret"]

        if not is_valid_client_secret(clientSecret):
            raise MatrixRestError(
                400, "M_INVALID_PARAM", "Invalid client_secret provided"
            )

        if account:
            # This is a v2 API so only allow binding to the logged in user id
            if account.userId != mxid:
                raise MatrixRestError(
                    403,
                    "M_UNAUTHORIZED",
                    "This user is prohibited from binding to the mxid",
                )

        try:
            valSessionStore = ThreePidValSessionStore(self.sydent)
            s = valSessionStore.getValidatedSession(sid, clientSecret)
        except (IncorrectClientSecretException, InvalidSessionIdException):
            # Return the same error for not found / bad client secret otherwise
            # people can get information about sessions without knowing the
            # secret.
            raise MatrixRestError(
                404,
                "M_NO_VALID_SESSION",
                "No valid session was found matching that sid and client secret",
            )
        except SessionExpiredException:
            raise MatrixRestError(
                400,
                "M_SESSION_EXPIRED",
                "This validation session has expired: call requestToken again",
            )
        except SessionNotValidatedException:
            raise MatrixRestError(
                400,
                "M_SESSION_NOT_VALIDATED",
                "This validation session has not yet been completed",
            )

        res = self.sydent.threepidBinder.addBinding(s.medium, s.address, mxid)
        return res

    def render_OPTIONS(self, request: Request) -> bytes:
        send_cors(request)
        return b""



import logging
from typing import TYPE_CHECKING, Dict

from sydent.db.valsession import ThreePidValSessionStore
from sydent.util import time_msec
from sydent.validators import (
    THREEPID_SESSION_VALIDATION_TIMEOUT_MS,
    IncorrectClientSecretException,
    IncorrectSessionTokenException,
    InvalidSessionIdException,
    SessionExpiredException,
)

if TYPE_CHECKING:
    from sydent.sydent import Sydent

logger = logging.getLogger(__name__)


def validateSessionWithToken(
    sydent: "Sydent", sid: int, clientSecret: str, token: str
) -> Dict[str, bool]:
    """
    Attempt to validate a session, identified by the sid, using
    the token from out-of-band. The client secret is given to
    prevent attempts to guess the token for a sid.

    :param sid: The ID of the session to validate.
    :param clientSecret: The client secret to validate.
    :param token: The token to validate.

    :return: A dict with a "success" key which is True if the session
        was successfully validated, False otherwise.

    :raise IncorrectClientSecretException: The provided client_secret is incorrect.
    :raise SessionExpiredException: The session has expired.
    :raise InvalidSessionIdException: The session ID couldn't be matched with an
        existing session.
    :raise IncorrectSessionTokenException: The provided token is incorrect
    """
    valSessionStore = ThreePidValSessionStore(sydent)
    result = valSessionStore.getTokenSessionById(sid)
    if not result:
        logger.info("Session ID %s not found", sid)
        raise InvalidSessionIdException()

    session, token_info = result

    if not clientSecret == session.client_secret:
        logger.info("Incorrect client secret", sid)
        raise IncorrectClientSecretException()

    if session.mtime + THREEPID_SESSION_VALIDATION_TIMEOUT_MS < time_msec():
        logger.info("Session expired")
        raise SessionExpiredException()

    # TODO once we can validate the token oob
    # if tokenObj.validated and clientSecret == tokenObj.clientSecret:
    #    return True

    if token_info.token == token:
        logger.info("Setting session %s as validated", session.id)
        valSessionStore.setValidated(session.id, True)

        return {"success": True}
    else:
        logger.info("Incorrect token submitted")
        raise IncorrectSessionTokenException()



# Copyright 2014 OpenMarket Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
from io import BytesIO
from typing import TYPE_CHECKING, Optional

from twisted.internet.defer import Deferred
from twisted.internet.interfaces import IOpenSSLClientConnectionCreator
from twisted.internet.ssl import optionsForClientTLS
from twisted.web.client import Agent, FileBodyProducer, Response
from twisted.web.http_headers import Headers
from twisted.web.iweb import IPolicyForHTTPS
from zope.interface import implementer

from sydent.types import JsonDict

if TYPE_CHECKING:
    from sydent.sydent import Sydent

logger = logging.getLogger(__name__)


class ReplicationHttpsClient:
    """
    An HTTPS client specifically for talking replication to other Matrix Identity Servers
    (ie. presents our replication SSL certificate and validates peer SSL certificates as we would in the
    replication HTTPS server)
    """

    def __init__(self, sydent: "Sydent") -> None:
        self.sydent = sydent
        self.agent: Optional[Agent] = None

        if self.sydent.sslComponents.myPrivateCertificate:
            # We will already have logged a warn if this is absent, so don't do it again
            # cert = self.sydent.sslComponents.myPrivateCertificate
            # self.certOptions = twisted.internet.ssl.CertificateOptions(privateKey=cert.privateKey.original,
            #                                                      certificate=cert.original,
            #                                                      trustRoot=self.sydent.sslComponents.trustRoot)
            self.agent = Agent(self.sydent.reactor, SydentPolicyForHTTPS(self.sydent))

    def postJson(
        self, uri: str, jsonObject: JsonDict
    ) -> Optional["Deferred[Response]"]:
        """
        Sends an POST request over HTTPS.

        :param uri: The URI to send the request to.
        :param jsonObject: The request's body.

        :return: The request's response.
        """
        logger.debug("POSTing request to %s", uri)
        if not self.agent:
            logger.error("HTTPS post attempted but HTTPS is not configured")
            return None

        headers = Headers(
            {"Content-Type": ["application/json"], "User-Agent": ["Sydent"]}
        )

        json_bytes = json.dumps(jsonObject).encode("utf8")
        reqDeferred = self.agent.request(
            b"POST", uri.encode("utf8"), headers, FileBodyProducer(BytesIO(json_bytes))
        )

        return reqDeferred


@implementer(IPolicyForHTTPS)
class SydentPolicyForHTTPS:
    def __init__(self, sydent: "Sydent") -> None:
        self.sydent = sydent

    def creatorForNetloc(
        self, hostname: bytes, port: int
    ) -> IOpenSSLClientConnectionCreator:
        return optionsForClientTLS(
            hostname.decode("ascii"),
            trustRoot=self.sydent.sslComponents.trustRoot,
            clientCertificate=self.sydent.sslComponents.myPrivateCertificate,
        )



# Copyright 2021 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from configparser import ConfigParser
from typing import List

from jinja2.environment import Environment
from jinja2.loaders import FileSystemLoader

from sydent.config._base import BaseConfig
from sydent.util.ip_range import DEFAULT_IP_RANGE_BLACKLIST, generate_ip_set


class GeneralConfig(BaseConfig):
    def parse_config(self, cfg: "ConfigParser") -> bool:
        """
        Parse the 'general' section of the config

        :param cfg: the configuration to be parsed
        """
        self.server_name = cfg.get("general", "server.name")
        if self.server_name == "":
            self.server_name = os.uname()[1]
            print(
                "WARNING: You have not specified a server name. I have guessed that this "
                f"server is called '{self.server_name}'. If this is incorrect, you should "
                "edit 'general.server.name' in the config file."
            )

        self.log_level = cfg.get("general", "log.level")
        self.log_path = cfg.get("general", "log.path")

        # Get the possible brands by looking at directories under the
        # templates.path directory.
        self.templates_path = cfg.get("general", "templates.path")
        if os.path.exists(self.templates_path):
            self.valid_brands = {
                p
                for p in os.listdir(self.templates_path)
                if os.path.isdir(os.path.join(self.templates_path, p))
            }
        else:
            print(
                f"WARNING: The path specified by 'general.templates.path' ({self.templates_path}) "
                "does not exist."
            )
            # This is a legacy code-path and assumes that verify_response_template,
            # email.template, and email.invite_template are defined.
            self.valid_brands = set()

        self.template_environment = Environment(
            loader=FileSystemLoader(cfg.get("general", "templates.path")),
            autoescape=True,
        )

        self.default_brand = cfg.get("general", "brand.default")

        self.pidfile = cfg.get("general", "pidfile.path")

        self.terms_path = cfg.get("general", "terms.path")

        self.address_lookup_limit = cfg.getint("general", "address_lookup_limit")

        self.prometheus_port = cfg.getint("general", "prometheus_port", fallback=None)
        self.prometheus_addr = cfg.get("general", "prometheus_addr", fallback=None)
        self.prometheus_enabled = (
            self.prometheus_port is not None and self.prometheus_addr is not None
        )

        self.sentry_enabled = cfg.has_option("general", "sentry_dsn")
        self.sentry_dsn = cfg.get("general", "sentry_dsn", fallback=None)

        self.delete_tokens_on_bind = parse_cfg_bool(
            cfg.get("general", "delete_tokens_on_bind")
        )

        ip_blacklist = list_from_comma_sep_string(cfg.get("general", "ip.blacklist"))
        if not ip_blacklist:
            ip_blacklist = DEFAULT_IP_RANGE_BLACKLIST

        ip_whitelist = list_from_comma_sep_string(cfg.get("general", "ip.whitelist"))

        self.ip_blacklist = generate_ip_set(ip_blacklist)
        self.ip_whitelist = generate_ip_set(ip_whitelist)

        self.enable_v1_access = parse_cfg_bool(cfg.get("general", "enable_v1_access"))

        homeserver_allow_list = list_from_comma_sep_string(
            cfg.get("general", "homeserver_allow_list")
        )
        if homeserver_allow_list and self.enable_v1_access:
            raise RuntimeError(
                """The V1 api must be disabled for the `homeserver_allow_list` to function, if you have 
                specified a `homeserver_allow_list` in the config file please ensure that the config 
                option `enable_v1_access` is set to 'false'."""
            )
        self.homeserver_allow_list = homeserver_allow_list

        if not self.enable_v1_access:
            self.enable_v1_associations = False
        else:
            self.enable_v1_associations = parse_cfg_bool(
                cfg.get("general", "enable_v1_associations")
            )

        return False


def list_from_comma_sep_string(rawstr: str) -> List[str]:
    """
    Parse the a comma seperated string into a list

    :param rawstr: the string to be parsed
    """
    if rawstr == "":
        return []
    return [x.strip() for x in rawstr.split(",")]


def parse_cfg_bool(value: str) -> bool:
    """
    Parse a string config option into a boolean
    This method ignores capitalisation

    :param value: the string to be parsed
    """
    return value.lower() == "true"



# Copyright 2014 OpenMarket Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Dict, Optional

import attr


def threePidAssocFromDict(d: Dict[str, Any]) -> "ThreepidAssociation":
    """Instantiates a ThreepidAssociation from the given dict."""
    assoc = ThreepidAssociation(
        d["medium"],
        d["address"],
        None,  # empty lookup_hash digest by default
        d["mxid"],
        d["ts"],
        d["not_before"],
        d["not_after"],
    )
    return assoc


@attr.s(slots=True, auto_attribs=True)
class ThreepidAssociation:
    """
    medium: The medium of the 3pid (eg. email)
    address: The identifier (eg. email address)
    lookup_hash: A hash digest of the 3pid. Can be a str or None
    mxid: The matrix ID the 3pid is associated with
    ts: The creation timestamp of this association, ms
    not_before: The timestamp, in ms, at which this association becomes valid
    not_after: The timestamp, in ms, at which this association ceases to be valid
    """

    medium: str
    address: str
    lookup_hash: Optional[str]
    # Note: the next four fields were made optional in schema version 2.
    # See sydent.db.sqlitedb.SqliteDatabase._upgradeSchema
    mxid: Optional[str]
    ts: Optional[int]
    not_before: Optional[int]
    not_after: Optional[int]
    extra_fields: Dict[str, Any] = {}



# Copyright 2014 OpenMarket Ltd
# Copyright 2018 New Vector Ltd
# Copyright 2019 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gc
import logging
import logging.handlers
import os
import sqlite3
from typing import Optional

import attr
import prometheus_client
import twisted.internet.reactor
from matrix_common.versionstring import get_distribution_version_string
from signedjson.types import SigningKey
from twisted.internet import address, task
from twisted.internet.interfaces import (
    IReactorCore,
    IReactorPluggableNameResolver,
    IReactorSSL,
    IReactorTCP,
    IReactorTime,
)
from twisted.python import log
from twisted.web.http import Request
from zope.interface import Interface

from sydent.config import SydentConfig
from sydent.db.hashing_metadata import HashingMetadataStore
from sydent.db.sqlitedb import SqliteDatabase
from sydent.db.valsession import ThreePidValSessionStore
from sydent.hs_federation.verifier import Verifier
from sydent.http.httpcommon import SslComponents
from sydent.http.httpsclient import ReplicationHttpsClient
from sydent.http.httpserver import (
    ClientApiHttpServer,
    InternalApiHttpServer,
    ReplicationHttpsServer,
)
from sydent.replication.pusher import Pusher
from sydent.threepid.bind import ThreepidBinder
from sydent.util.hash import sha256_and_url_safe_base64
from sydent.util.ratelimiter import Ratelimiter
from sydent.util.tokenutils import generateAlphanumericTokenOfLength
from sydent.validators.emailvalidator import EmailValidator
from sydent.validators.msisdnvalidator import MsisdnValidator

logger = logging.getLogger(__name__)


class SydentReactor(
    IReactorCore,
    IReactorTCP,
    IReactorSSL,
    IReactorTime,
    IReactorPluggableNameResolver,
    Interface,
):
    pass


class Sydent:
    def __init__(
        self,
        sydent_config: SydentConfig,
        reactor: SydentReactor = twisted.internet.reactor,  # type: ignore[assignment]
        use_tls_for_federation: bool = True,
    ):
        self.config = sydent_config

        self.reactor = reactor
        self.use_tls_for_federation = use_tls_for_federation

        logger.info("Starting Sydent server")

        self.db: sqlite3.Connection = SqliteDatabase(self).db

        if self.config.general.sentry_enabled:
            import sentry_sdk

            sentry_sdk.init(
                dsn=self.config.general.sentry_dsn,
                release=get_distribution_version_string("matrix-sydent"),
            )
            with sentry_sdk.configure_scope() as scope:
                scope.set_tag("sydent_server_name", self.config.general.server_name)

            # workaround for https://github.com/getsentry/sentry-python/issues/803: we
            # disable automatic GC and run it periodically instead.
            gc.disable()
            cb = task.LoopingCall(run_gc)
            cb.clock = self.reactor
            cb.start(1.0)

        # See if a pepper already exists in the database
        # Note: This MUST be run before we start serving requests, otherwise lookups for
        # 3PID hashes may come in before we've completed generating them
        hashing_metadata_store = HashingMetadataStore(self)
        lookup_pepper = hashing_metadata_store.get_lookup_pepper()
        if not lookup_pepper:
            # No pepper defined in the database, generate one
            lookup_pepper = generateAlphanumericTokenOfLength(5)

            # Store it in the database and rehash 3PIDs
            hashing_metadata_store.store_lookup_pepper(
                sha256_and_url_safe_base64, lookup_pepper
            )

        self.validators: Validators = Validators(
            EmailValidator(self), MsisdnValidator(self)
        )

        self.keyring: Keyring = Keyring(self.config.crypto.signing_key)
        self.keyring.ed25519.alg = "ed25519"

        self.sig_verifier: Verifier = Verifier(self)

        self.threepidBinder: ThreepidBinder = ThreepidBinder(self)

        self.sslComponents: SslComponents = SslComponents(self)

        self.clientApiHttpServer = ClientApiHttpServer(self, lookup_pepper)
        self.replicationHttpsServer = ReplicationHttpsServer(self)
        self.replicationHttpsClient: ReplicationHttpsClient = ReplicationHttpsClient(
            self
        )

        self.pusher: Pusher = Pusher(self)

        self.email_sender_ratelimiter: Ratelimiter[str] = Ratelimiter(
            self.reactor,
            burst=self.config.email.email_sender_ratelimit_burst,
            rate_hz=self.config.email.email_sender_ratelimit_rate_hz,
        )

    def run(self) -> None:
        self.clientApiHttpServer.setup()
        self.replicationHttpsServer.setup()
        self.pusher.setup()
        self.maybe_start_prometheus_server()

        # A dedicated validation session store just to clean up old sessions every N minutes
        self.cleanupValSession = ThreePidValSessionStore(self)
        cb = task.LoopingCall(self.cleanupValSession.deleteOldSessions)
        cb.clock = self.reactor
        cb.start(10 * 60.0)

        if self.config.http.internal_port is not None:
            internalport = self.config.http.internal_port
            interface = self.config.http.internal_bind_address

            self.internalApiHttpServer = InternalApiHttpServer(self)
            self.internalApiHttpServer.setup(interface, internalport)

        if self.config.general.pidfile:
            with open(self.config.general.pidfile, "w") as pidfile:
                pidfile.write(str(os.getpid()) + "\n")

        self.reactor.run()

    def maybe_start_prometheus_server(self) -> None:
        if self.config.general.prometheus_enabled:
            assert self.config.general.prometheus_addr is not None
            assert self.config.general.prometheus_port is not None
            prometheus_client.start_http_server(
                port=self.config.general.prometheus_port,
                addr=self.config.general.prometheus_addr,
            )

    def ip_from_request(self, request: Request) -> Optional[str]:
        if self.config.http.obey_x_forwarded_for and request.requestHeaders.hasHeader(
            "X-Forwarded-For"
        ):
            # Type safety: hasHeaders returning True means that getRawHeaders
            # returns a nonempty list
            return request.requestHeaders.getRawHeaders("X-Forwarded-For")[0]  # type: ignore[index]
        client = request.getClientAddress()
        if isinstance(client, (address.IPv4Address, address.IPv6Address)):
            return client.host
        else:
            return None

    def brand_from_request(self, request: Request) -> Optional[str]:
        """
        If the brand GET parameter is passed, returns that as a string, otherwise returns None.

        :param request: The incoming request.

        :return: The brand to use or None if no hint is found.
        """
        if b"brand" in request.args:
            return request.args[b"brand"][0].decode("utf-8")
        return None

    def get_branded_template(
        self,
        brand: Optional[str],
        template_name: str,
    ) -> str:
        """
        Calculate a branded template filename to use.

        Attempt to use the hinted brand from the request if the brand
        is valid. Otherwise, fallback to the default brand.

        :param brand: The hint of which brand to use.
        :type brand: str or None
        :param template_name: The name of the template file to load.
        :type template_name: str

        :return: The template filename to use.
        :rtype: str
        """

        # If a brand hint is provided, attempt to use it if it is valid.
        if brand:
            if brand not in self.config.general.valid_brands:
                brand = None

        # If the brand hint is not valid, or not provided, fallback to the default brand.
        if not brand:
            brand = self.config.general.default_brand

        root_template_path = self.config.general.templates_path

        # Grab jinja template if it exists
        if os.path.exists(
            os.path.join(root_template_path, brand, template_name + ".j2")
        ):
            return os.path.join(brand, template_name + ".j2")
        else:
            return os.path.join(root_template_path, brand, template_name)


@attr.s(frozen=True, slots=True, auto_attribs=True)
class Validators:
    email: EmailValidator
    msisdn: MsisdnValidator


@attr.s(frozen=True, slots=True, auto_attribs=True)
class Keyring:
    ed25519: SigningKey


def get_config_file_path() -> str:
    return os.environ.get("SYDENT_CONF", "sydent.conf")


def run_gc() -> None:
    threshold = gc.get_threshold()
    counts = gc.get_count()
    for i in reversed(range(len(threshold))):
        if threshold[i] < counts[i]:
            gc.collect(i)


def setup_logging(config: SydentConfig) -> None:
    """
    Setup logging using the options specified in the config

    :param config: the configuration to use
    """
    log_path = config.general.log_path
    log_level = config.general.log_level

    log_format = "%(asctime)s - %(name)s - %(lineno)d - %(levelname)s" " - %(message)s"
    formatter = logging.Formatter(log_format)

    handler: logging.Handler
    if log_path != "":
        handler = logging.handlers.TimedRotatingFileHandler(
            log_path, when="midnight", backupCount=365
        )
        handler.setFormatter(formatter)

    else:
        handler = logging.StreamHandler()

    handler.setFormatter(formatter)
    rootLogger = logging.getLogger("")
    rootLogger.setLevel(log_level)
    rootLogger.addHandler(handler)

    observer = log.PythonLoggingObserver()
    observer.start()


def main() -> None:
    sydent_config = SydentConfig()
    sydent_config.parse_config_file(get_config_file_path())
    setup_logging(sydent_config)

    syd = Sydent(sydent_config)
    syd.run()


if __name__ == "__main__":
    main()



# Copyright 2014 OpenMarket Ltd
# Copyright 2018 New Vector Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from http import HTTPStatus
from typing import TYPE_CHECKING

from signedjson.sign import SignatureVerifyException
from twisted.internet import defer
from twisted.internet.error import ConnectError, DNSLookupError
from twisted.web import server
from twisted.web.client import ResponseFailed
from twisted.web.server import Request

from sydent.db.valsession import ThreePidValSessionStore
from sydent.hs_federation.verifier import InvalidServerName, NoAuthenticationError
from sydent.http.servlets import SydentResource, dict_to_json_bytes
from sydent.types import JsonDict
from sydent.util import json_decoder
from sydent.util.stringutils import is_valid_client_secret
from sydent.validators import (
    IncorrectClientSecretException,
    InvalidSessionIdException,
    SessionNotValidatedException,
)

if TYPE_CHECKING:
    from sydent.sydent import Sydent

logger = logging.getLogger(__name__)


class ThreePidUnbindServlet(SydentResource):
    def __init__(self, sydent: "Sydent") -> None:
        super().__init__()
        self.sydent = sydent

    def render_POST(
        self, request: Request
    ) -> object:  # from the twisted docs: @type NOT_DONE_YET is an opaque object
        defer.ensureDeferred(self._async_render_POST(request))
        return server.NOT_DONE_YET

    async def _async_render_POST(self, request: Request) -> None:
        try:
            try:
                # TODO: we should really validate that this gives us a dict, and
                #   not some other json value like str, list, int etc
                # json.loads doesn't allow bytes in Python 3.5
                body: JsonDict = json_decoder.decode(
                    request.content.read().decode("UTF-8")
                )
            except ValueError:
                request.setResponseCode(HTTPStatus.BAD_REQUEST)
                request.write(
                    dict_to_json_bytes(
                        {"errcode": "M_BAD_JSON", "error": "Malformed JSON"}
                    )
                )
                request.finish()
                return

            missing = [k for k in ("threepid", "mxid") if k not in body]
            if len(missing) > 0:
                request.setResponseCode(HTTPStatus.BAD_REQUEST)
                msg = "Missing parameters: " + (",".join(missing))
                request.write(
                    dict_to_json_bytes({"errcode": "M_MISSING_PARAMS", "error": msg})
                )
                request.finish()
                return

            threepid = body["threepid"]
            mxid = body["mxid"]

            if "medium" not in threepid or "address" not in threepid:
                request.setResponseCode(HTTPStatus.BAD_REQUEST)
                request.write(
                    dict_to_json_bytes(
                        {
                            "errcode": "M_MISSING_PARAMS",
                            "error": "Threepid lacks medium / address",
                        }
                    )
                )
                request.finish()
                return

            # We now check for authentication in two different ways, depending
            # on the contents of the request. If the user has supplied "sid"
            # (the Session ID returned by Sydent during the original binding)
            # and "client_secret" fields, they are trying to prove that they
            # were the original author of the bind. We then check that what
            # they supply matches and if it does, allow the unbind.
            #
            # However if these fields are not supplied, we instead check
            # whether the request originated from a homeserver, and if so the
            # same homeserver that originally created the bind. We do this by
            # checking the signature of the request. If it all matches up, we
            # allow the unbind.
            #
            # Only one method of authentication is required.
            if "sid" in body and "client_secret" in body:
                sid = body["sid"]
                client_secret = body["client_secret"]

                if not is_valid_client_secret(client_secret):
                    request.setResponseCode(HTTPStatus.BAD_REQUEST)
                    request.write(
                        dict_to_json_bytes(
                            {
                                "errcode": "M_INVALID_PARAM",
                                "error": "Invalid client_secret provided",
                            }
                        )
                    )
                    request.finish()
                    return

                valSessionStore = ThreePidValSessionStore(self.sydent)

                try:
                    s = valSessionStore.getValidatedSession(sid, client_secret)
                except (IncorrectClientSecretException, InvalidSessionIdException):
                    request.setResponseCode(HTTPStatus.UNAUTHORIZED)
                    request.write(
                        dict_to_json_bytes(
                            {
                                "errcode": "M_NO_VALID_SESSION",
                                "error": "No valid session was found matching that sid and client secret",
                            }
                        )
                    )
                    request.finish()
                    return
                except SessionNotValidatedException:
                    request.setResponseCode(HTTPStatus.FORBIDDEN)
                    request.write(
                        dict_to_json_bytes(
                            {
                                "errcode": "M_SESSION_NOT_VALIDATED",
                                "error": "This validation session has not yet been completed",
                            }
                        )
                    )
                    return

                if s.medium != threepid["medium"] or s.address != threepid["address"]:
                    request.setResponseCode(HTTPStatus.FORBIDDEN)
                    request.write(
                        dict_to_json_bytes(
                            {
                                "errcode": "M_FORBIDDEN",
                                "error": "Provided session information does not match medium/address combo",
                            }
                        )
                    )
                    request.finish()
                    return
            else:
                try:
                    origin_server_name = (
                        await self.sydent.sig_verifier.authenticate_request(
                            request, body
                        )
                    )
                except SignatureVerifyException as ex:
                    request.setResponseCode(HTTPStatus.UNAUTHORIZED)
                    request.write(
                        dict_to_json_bytes({"errcode": "M_FORBIDDEN", "error": str(ex)})
                    )
                    request.finish()
                    return
                except NoAuthenticationError as ex:
                    request.setResponseCode(HTTPStatus.UNAUTHORIZED)
                    request.write(
                        dict_to_json_bytes({"errcode": "M_FORBIDDEN", "error": str(ex)})
                    )
                    request.finish()
                    return
                except InvalidServerName as ex:
                    request.setResponseCode(HTTPStatus.BAD_REQUEST)
                    request.write(
                        dict_to_json_bytes(
                            {"errcode": "M_INVALID_PARAM", "error": str(ex)}
                        )
                    )
                    request.finish()
                    return
                except (DNSLookupError, ConnectError, ResponseFailed) as e:
                    msg = (
                        f"Unable to contact the Matrix homeserver to "
                        f"authenticate request ({type(e).__name__})"
                    )
                    logger.warning(msg)
                    request.setResponseCode(HTTPStatus.INTERNAL_SERVER_ERROR)
                    request.write(
                        dict_to_json_bytes(
                            {
                                "errcode": "M_UNKNOWN",
                                "error": msg,
                            }
                        )
                    )
                    request.finish()
                    return
                except Exception:
                    logger.exception("Exception whilst authenticating unbind request")
                    request.setResponseCode(HTTPStatus.INTERNAL_SERVER_ERROR)
                    request.write(
                        dict_to_json_bytes(
                            {"errcode": "M_UNKNOWN", "error": "Internal Server Error"}
                        )
                    )
                    request.finish()
                    return

                if not mxid.endswith(":" + origin_server_name):
                    request.setResponseCode(HTTPStatus.FORBIDDEN)
                    request.write(
                        dict_to_json_bytes(
                            {
                                "errcode": "M_FORBIDDEN",
                                "error": "Origin server name does not match mxid",
                            }
                        )
                    )
                    request.finish()
                    return

            self.sydent.threepidBinder.removeBinding(threepid, mxid)

            request.write(dict_to_json_bytes({}))
            request.finish()
        except Exception as ex:
            logger.exception("Exception whilst handling unbind")
            request.setResponseCode(HTTPStatus.INTERNAL_SERVER_ERROR)
            request.write(
                dict_to_json_bytes({"errcode": "M_UNKNOWN", "error": str(ex)})
            )
            request.finish()



# Copyright 2022 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from twisted.web.server import Request

from sydent.http.servlets import SydentResource, jsonwrap, send_cors
from sydent.types import JsonDict


class VersionsServlet(SydentResource):
    isLeaf = True

    @jsonwrap
    def render_GET(self, request: Request) -> JsonDict:
        """
        Return the supported Matrix versions.
        """
        send_cors(request)

        return {
            "versions": [
                "r0.1.0",
                "r0.2.0",
                "r0.2.1",
                "r0.3.0",
                "v1.1",
                "v1.2",
                "v1.3",
                "v1.4",
                "v1.5",
            ]
        }

    def render_OPTIONS(self, request: Request) -> bytes:
        send_cors(request)
        return b""



# Copyright 2016 OpenMarket Ltd
# Copyright 2017 Vector Creations Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import TYPE_CHECKING

import phonenumbers
from twisted.web.server import Request

from sydent.http.auth import authV2
from sydent.http.servlets import (
    SydentResource,
    asyncjsonwrap,
    get_args,
    jsonwrap,
    send_cors,
)
from sydent.types import JsonDict
from sydent.util.ratelimiter import Ratelimiter
from sydent.util.stringutils import is_valid_client_secret
from sydent.validators import (
    DestinationRejectedException,
    IncorrectClientSecretException,
    IncorrectSessionTokenException,
    InvalidSessionIdException,
    SessionExpiredException,
)

if TYPE_CHECKING:
    from sydent.sydent import Sydent

logger = logging.getLogger(__name__)


class MsisdnRequestCodeServlet(SydentResource):
    isLeaf = True

    def __init__(self, syd: "Sydent", require_auth: bool = False) -> None:
        super().__init__()
        self.sydent = syd
        self.require_auth = require_auth
        self._msisdn_ratelimiter = Ratelimiter[str](
            syd.reactor,
            syd.config.sms.msisdn_ratelimit_burst,
            syd.config.sms.msisdn_ratelimit_rate_hz,
        )
        self._country_ratelimiter = Ratelimiter[int](
            syd.reactor,
            syd.config.sms.country_ratelimit_burst,
            syd.config.sms.country_ratelimit_rate_hz,
        )

    @asyncjsonwrap
    async def render_POST(self, request: Request) -> JsonDict:
        send_cors(request)

        if self.require_auth:
            authV2(self.sydent, request)

        args = get_args(
            request, ("phone_number", "country", "client_secret", "send_attempt")
        )

        raw_phone_number = args["phone_number"]
        country = args["country"]
        try:
            # See the comment handling `send_attempt` in emailservlet.py for
            # more context.
            sendAttempt = int(args["send_attempt"])
        except (TypeError, ValueError):
            request.setResponseCode(400)
            return {
                "errcode": "M_INVALID_PARAM",
                "error": f"send_attempt should be an integer (got {args['send_attempt']}",
            }
        clientSecret = args["client_secret"]

        if not is_valid_client_secret(clientSecret):
            request.setResponseCode(400)
            return {
                "errcode": "M_INVALID_PARAM",
                "error": "Invalid client_secret provided",
            }

        try:
            phone_number_object = phonenumbers.parse(raw_phone_number, country)

            if phone_number_object.country_code is None:
                raise Exception("No country code")
        except Exception as e:
            logger.warning("Invalid phone number given: %r", e)
            request.setResponseCode(400)
            return {
                "errcode": "M_INVALID_PHONE_NUMBER",
                "error": "Invalid phone number",
            }

        msisdn = phonenumbers.format_number(
            phone_number_object, phonenumbers.PhoneNumberFormat.E164
        )[1:]

        self._msisdn_ratelimiter.ratelimit(msisdn, "Limit exceeded for this number")
        self._country_ratelimiter.ratelimit(
            phone_number_object.country_code, "Limit exceeded for this country"
        )

        # International formatted number. The same as an E164 but with spaces
        # in appropriate places to make it nicer for the humans.
        intl_fmt = phonenumbers.format_number(
            phone_number_object, phonenumbers.PhoneNumberFormat.INTERNATIONAL
        )

        brand = self.sydent.brand_from_request(request)
        try:
            sid = await self.sydent.validators.msisdn.requestToken(
                phone_number_object, clientSecret, sendAttempt, brand
            )
            resp = {
                "success": True,
                "sid": str(sid),
                "msisdn": msisdn,
                "intl_fmt": intl_fmt,
            }
        except DestinationRejectedException:
            logger.warning("Destination rejected for number: %s", msisdn)
            request.setResponseCode(400)
            resp = {
                "errcode": "M_DESTINATION_REJECTED",
                "error": "Phone numbers in this country are not currently supported",
            }
        except Exception:
            logger.exception("Exception sending SMS")
            request.setResponseCode(500)
            resp = {"errcode": "M_UNKNOWN", "error": "Internal Server Error"}

        return resp

    def render_OPTIONS(self, request: Request) -> bytes:
        send_cors(request)
        return b""


class MsisdnValidateCodeServlet(SydentResource):
    isLeaf = True

    def __init__(self, syd: "Sydent", require_auth: bool = False) -> None:
        super().__init__()
        self.sydent = syd
        self.require_auth = require_auth

    def render_GET(self, request: Request) -> str:
        send_cors(request)

        args = get_args(request, ("token", "sid", "client_secret"))
        resp = self.do_validate_request(request)
        if "success" in resp and resp["success"]:
            msg = "Verification successful! Please return to your Matrix client to continue."
            if "next_link" in args:
                next_link = args["next_link"]
                request.setResponseCode(302)
                request.setHeader("Location", next_link)
        else:
            request.setResponseCode(400)
            msg = (
                "Verification failed: you may need to request another verification text"
            )

        brand = self.sydent.brand_from_request(request)

        # self.sydent.config.http.verify_response_template is deprecated
        if self.sydent.config.http.verify_response_template is None:
            templateFile = self.sydent.get_branded_template(
                brand,
                "verify_response_template.html",
            )
        else:
            templateFile = self.sydent.config.http.verify_response_template

        request.setHeader("Content-Type", "text/html")
        return open(templateFile).read() % {"message": msg}

    @jsonwrap
    def render_POST(self, request: Request) -> JsonDict:
        send_cors(request)

        if self.require_auth:
            authV2(self.sydent, request)

        return self.do_validate_request(request)

    def do_validate_request(self, request: Request) -> JsonDict:
        """
        Extracts information about a validation session from the request and
        attempts to validate that session.

        :param request: The request to extract information about the session from.

        :return: A dict with a "success" key which value indicates whether the
            validation succeeded. If the validation failed, this dict also includes
            a "errcode" and a "error" keys which include information about the failure.
        """

        args = get_args(request, ("token", "sid", "client_secret"))

        sid = args["sid"]
        tokenString = args["token"]
        clientSecret = args["client_secret"]

        if not is_valid_client_secret(clientSecret):
            request.setResponseCode(400)
            return {
                "errcode": "M_INVALID_PARAM",
                "error": "Invalid client_secret provided",
            }

        try:
            return self.sydent.validators.msisdn.validateSessionWithToken(
                sid, clientSecret, tokenString
            )
        except IncorrectClientSecretException:
            request.setResponseCode(400)
            return {
                "success": False,
                "errcode": "M_INVALID_PARAM",
                "error": "Client secret does not match the one given when requesting the token",
            }
        except SessionExpiredException:
            request.setResponseCode(400)
            return {
                "success": False,
                "errcode": "M_SESSION_EXPIRED",
                "error": "This validation session has expired: call requestToken again",
            }
        except InvalidSessionIdException:
            request.setResponseCode(400)
            return {
                "success": False,
                "errcode": "M_INVALID_PARAM",
                "error": "The token doesn't match",
            }
        except IncorrectSessionTokenException:
            request.setResponseCode(404)
            return {
                "success": False,
                "errcode": "M_NO_VALID_SESSION",
                "error": "No session could be found with this sid",
            }

    def render_OPTIONS(self, request: Request) -> bytes:
        send_cors(request)
        return b""



# Copyright 2014 OpenMarket Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import functools
import json
import logging
from typing import Any, Awaitable, Callable, Dict, Iterable, TypeVar

from prometheus_client import Counter
from twisted.internet import defer
from twisted.web import server
from twisted.web.resource import Resource
from twisted.web.server import Request

from sydent.types import JsonDict
from sydent.util import json_decoder

logger = logging.getLogger(__name__)


request_counter = Counter(
    "sydent_http_received_requests",
    "Received requests",
    labelnames=("servlet", "method"),
)


class SydentResource(Resource):
    """A subclass of resource that tracks request metrics"""

    def __init__(self) -> None:
        self._name = self.__class__.__name__
        super().__init__()

    def render(self, request: Request) -> Any:
        request_counter.labels(self._name, request.method).inc()
        return super().render(request)


class MatrixRestError(Exception):
    """
    Handled by the jsonwrap wrapper. Any servlets that don't use this
    wrapper should catch this exception themselves.
    """

    def __init__(self, httpStatus: int, errcode: str, error: str):
        super(Exception, self).__init__(error)
        self.httpStatus = httpStatus
        self.errcode = errcode
        self.error = error


def get_args(
    request: Request, args: Iterable[str], required: bool = True
) -> Dict[str, Any]:
    """
    Helper function to get arguments for an HTTP request.
    Currently takes args from the top level keys of a json object or
    www-form-urlencoded for backwards compatibility on v1 endpoints only.

    ⚠️ BEWARE ⚠. If a v1 request provides its args in urlencoded form (either in
    a POST body or as URL query parameters), then we'll return `Dict[str, str]`.
    The caller may need to interpret these strings as e.g. an `int`, `bool`, etc.
    Arguments given as a json body are processed with `json.JSONDecoder.decode`,
    and so are automatically deserialised to a Python type. The caller should
    still validate that these have the correct type!

    :param request: The request received by the servlet.
    :param args: The args to look for in the request's parameters.
    :param required: Whether to raise a MatrixRestError with 400
        M_MISSING_PARAMS if an argument is not found.

    :raises: MatrixRestError if required is True and a given parameter
        was not found in the request's query parameters.
    :raises: MatrixRestError if we the request body contains bad JSON.
    :raises: MatrixRestError if arguments are given in www-form-urlencodedquery
        form, and some argument name or value is not a valid UTF-8-encoded
        string.

    :return: A dict containing the requested args and their values. String values
        are of type unicode.
    """
    assert request.path is not None
    v1_path = request.path.startswith(b"/_matrix/identity/api/v1")

    request_args = None
    # for v1 paths, only look for json args if content type is json
    if request.method in (b"POST", b"PUT") and (
        not v1_path
        or (
            request.requestHeaders.hasHeader("Content-Type")
            # type safety: getRawHeaders() will return a nonempty list because
            # the hasHeader call has returned True.
            and request.requestHeaders.getRawHeaders("Content-Type")[0].startswith(  # type: ignore[index]
                "application/json"
            )
        )
    ):
        try:
            # json.loads doesn't allow bytes in Python 3.5
            request_args = json_decoder.decode(request.content.read().decode("UTF-8"))
        except ValueError:
            raise MatrixRestError(400, "M_BAD_JSON", "Malformed JSON")

    # If we didn't get anything from that, and it's a v1 api path, try the request args
    # (element-web's usage of the ed25519 sign servlet currently involves
    # sending the params in the query string with a json body of 'null')
    if request_args is None and (v1_path or request.method == b"GET"):
        request_args_bytes = copy.copy(request.args)
        # Twisted supplies everything as an array because it's valid to
        # supply the same params multiple times with www-form-urlencoded
        # params. This make it incompatible with the json object though,
        # so we need to convert one of them. Since this is the
        # backwards-compat option, we convert this one.
        request_args = {}
        for k, v in request_args_bytes.items():
            if isinstance(v, list) and len(v) == 1:
                try:
                    request_args[k.decode("UTF-8")] = v[0].decode("UTF-8")
                except UnicodeDecodeError:
                    # Get a version of the key that has non-UTF-8 characters replaced by
                    # their \xNN escape sequence so it doesn't raise another exception.
                    safe_k = k.decode("UTF-8", errors="backslashreplace")
                    raise MatrixRestError(
                        400,
                        "M_INVALID_PARAM",
                        "Parameter %s and its value must be valid UTF-8" % safe_k,
                    )

    elif request_args is None:
        request_args = {}

    if required:
        # Check for any missing arguments
        missing = []
        for a in args:
            if a not in request_args:
                missing.append(a)

        if len(missing) > 0:
            request.setResponseCode(400)
            msg = "Missing parameters: " + (",".join(missing))
            raise MatrixRestError(400, "M_MISSING_PARAMS", msg)

    return request_args


Res = TypeVar("Res", bound=Resource)


def jsonwrap(f: Callable[[Res, Request], JsonDict]) -> Callable[[Res, Request], bytes]:
    @functools.wraps(f)
    def inner(self: Res, request: Request) -> bytes:

        """
        Runs a web handler function with the given request and parameters, then
        converts its result into JSON and returns it. If an error happens, also sets
        the HTTP response code.

        :param self: The current object.
        :param request: The request to process.

        :return: The JSON payload to send as a response to the request.
        """
        try:
            request.setHeader("Content-Type", "application/json")
            return dict_to_json_bytes(f(self, request))
        except MatrixRestError as e:
            request.setResponseCode(e.httpStatus)
            return dict_to_json_bytes({"errcode": e.errcode, "error": e.error})
        except Exception:
            logger.exception("Exception processing request")
            request.setHeader("Content-Type", "application/json")
            request.setResponseCode(500)
            return dict_to_json_bytes(
                {
                    "errcode": "M_UNKNOWN",
                    "error": "Internal Server Error",
                }
            )

    return inner


AsyncRenderer = Callable[[Res, Request], Awaitable[JsonDict]]


def asyncjsonwrap(f: AsyncRenderer[Res]) -> Callable[[Res, Request], object]:
    async def render(f: AsyncRenderer[Res], self: Res, request: Request) -> None:
        request.setHeader("Content-Type", "application/json")
        try:
            result = await f(self, request)
            request.write(dict_to_json_bytes(result))
        except MatrixRestError as e:
            request.setResponseCode(e.httpStatus)
            request.write(dict_to_json_bytes({"errcode": e.errcode, "error": e.error}))
        except Exception:
            logger.exception("Request processing failed")
            request.setResponseCode(500)
            request.write(
                dict_to_json_bytes(
                    {"errcode": "M_UNKNOWN", "error": "Internal Server Error"}
                )
            )
        request.finish()

    @functools.wraps(f)
    def inner(self: Res, request: Request) -> object:
        """
        Runs an asynchronous web handler function with the given arguments.

        :param self: The servelet instance
        :param request: The request that `f` will serve.

        :return: An opaque object to tell the servlet that the response isn't ready yet
            and will come later.
        """
        defer.ensureDeferred(render(f, self, request))
        return server.NOT_DONE_YET

    return inner


def send_cors(request: Request) -> None:
    request.setHeader("Access-Control-Allow-Origin", "*")
    request.setHeader("Access-Control-Allow-Methods", "GET, POST, PUT, DELETE, OPTIONS")
    request.setHeader("Access-Control-Allow-Headers", "*")


def dict_to_json_bytes(content: JsonDict) -> bytes:
    """
    Converts a dict into JSON and encodes it to bytes.

    :return: The JSON bytes.
    """
    return json.dumps(content).encode("UTF-8")
